/*
   Open BCrypt - 128 bit multiplatform file security

   Provided for free under the BSD license.

   compile for Windows using MSVC:
      cl bcrypt.cpp

   compile on linux systems:
      g++ -o bcrypt bcrypt.cpp -static

   about the name:
      written in 1999 by Juergen Thumm, 
      during my time at Brokat Infosystems.
      has nothing to do with the linux bcrypt package.
      in case of conflicts you may rename it to "obcrypt".

   compile with -DWITH_SELFTEST to support an option -selftest.
   with this option, bcrypt runs itself many times with generated
   test data in varying sizes.

   history
   -------
      27-Jun-23   V1.7.2 rework of code to make it compilable
                  with 64 bit compilers.

      20-Jun-23   V1.7.1 In-Memory encrypt/decrypt support.
                  removed asserts, all done by returncode.
                  all in one single source code file.

      06-Mar-20   V1.7 with large file support.
                  Open BCrypt is now provided under BSD license.

      28-Feb-04   incremented max iterations to 500000.
                  2 minor changes to enable Linux compile with c++ std compiler
		  (scope problem with loop variables).
		  set version string to 1.6.0 R2.

      10-Sep-99   added: single-parameter mode with extension autodetect
      14-Sep-99   general rework. usable pseudo-GUI version.
      16-Sep-99   solaris,aix,hp/ux,os/390.
      17-Sep-99   random padding, bc4, sha inner and md5 outer hash.
      20-Sep-99   added BRND block with random noise.
      21-Sep-99   os/2, linux.
      22-Sep-99   added key iteration phase to break brute force.
      24-Sep-99   added md5 hash for file header.
      26-Sep-99   long filenames now even with windows explorer.
      27-Sep-99   long extensions by default, progress indicator.
                  general rework and restructuring.
      28-Sep-99   reuse of keyFromSecret for phase2Keys.
      30-Sep-99   passphrase character restriction. hashing mode.
      08-Oct-99   options f,q,q2,a.
      11-Oct-99   V 1.3.3 - first official release.
      15-Oct-99   V 1.3.4 - minor reworks. ".brx.txt" is now also decrypted.
      22-Oct-99   V 1.4.1 - no name guessing, always scan infile.
                  added -selftest functionality for automated stress tests.
      23-Oct-99   V 1.4.2 - no longer placing original filename
                  in unencrypted file header part.
      27-Oct-99   V 1.5.0 - added "-remove" function for cryptographically
                  safe file deletion.
      29-Oct-99   V 1.5.1 - added file size check, rejecting > 2 GB now.
                  full -rmrf functionality (windows only).
                  selftest now with random passphrase.
      02-Nov-99   V 1.5.2 - minor cleanup, final version.
      02-Dec-99   V 1.6.0 - security improvements:
                  added passphrase salt to block prehashed table attack.
                  slightly optimized mathHash.
                  fix: hashing of encrypted files by pressing ESC didn't work.
      27-Dec-99   same version, added BEOS.
*/

#define VERSTR "1.7.2"

#ifdef __BR_OS390__
  #define __HAVE_NO_ALLOCA
  #define _OPEN_THREADS 1
  #define _XOPEN_SOURCE_EXTENDED 1
  #define S_IREAD   00200
  #define S_IWRITE  00400
  #include <sys/ioctl.h>
  #include <arpa/inet.h>
  #include <netdb.h>
  #include <pthread.h>
#endif

#ifdef _WIN32
  #define __BR_NT__
#else
  #define __BR_IX__
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdint.h>
#include <ctype.h>
#include <assert.h>
#include <time.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef __BR_NT__
  #include <windows.h>
  // #include <eh.h>
  #include <sys/timeb.h>
  #include <time.h>
  #include <process.h>
  #define getpid() _getpid()
  #include <errno.h>
  #include <direct.h>
#elif defined(__BR_OS2__)
  #include <os2def.h>
  #include <conio.h>
  #define INCL_DOSPROCESS
  #include <bsedos.h>
  #include <process.h>
#else
  #include <sys/resource.h>
  #include <unistd.h>
  #include <netdb.h>
  #include <netinet/in.h>
  #include <sys/utsname.h>
  #include <sys/resource.h>
  #include <sys/time.h>
#endif

#ifdef __BR_SUN__
  #include <sys/procfs.h>
#endif

#ifdef _WIN32
 #if _MSC_VER >= 1310
  #define SFK_W64
 #endif
 #if _MSC_VER >= 1900 // visual c++ 14.0
  #define SFK_NO_MEMTRACE
  #define SFK_VC14
  #define sfkfinddata64_t _wfinddata64_t
 #else
  #define sfkfinddata64_t __wfinddata64_t
 #endif
#endif

#ifdef _WIN32
 typedef __int64 num;
 typedef unsigned __int64 unum;
 #ifdef SFK_W64
  typedef __time64_t mytime_t;
  #define mymktime _mktime64
  #define mytime _time64
 #else
  typedef time_t mytime_t;
  #define mymktime mktime
  #define mytime time
 #endif
#else
 typedef long long num;
 typedef unsigned long long unum;
 typedef time_t mytime_t;
 #define mymktime mktime
 #define mytime time
#endif

#ifdef _WIN32
   #ifdef _MSC_VER
      #ifdef SFK_W64
      typedef struct __stat64 sfkstat_t;
      #else
      typedef struct stat sfkstat_t;
      #endif
   #else
      typedef struct stat sfkstat_t;
   #endif
#else
   typedef struct stat64 sfkstat_t;
#endif

#define mymin(a,b) ((a<b)?(a):(b))

#define MIN_ITERATIONS 1500
// #define MEMTRACE

#define CIPHER_3DES  1  // three 56-bit DES
#define CIPHER_BC4   2  // 128-bit bc4
#define CIPHER_IDEA  3  // single 128-bit idea

#define WITH_BC4        // proven in browsers
// #define WITH_3DES    // too slow
// #define WITH_IDEA      // patented

#ifdef MEMTRACE

// memory tracing, inactive in release compile
static int32_t lClAllocs = 0;
static int32_t lClFrees  = 0;
static int32_t lMemTotal = 0;
static int32_t alSize[4096];
static unsigned char *apMem[4096];

static const char *pszGlblMemFile;
static int32_t lGlblMemLine;
static void setMemPos(const char *pszFile, int32_t line)
{  pszGlblMemFile = pszFile; lGlblMemLine = line; }

static void *operator new(size_t size, char *pszFile, int32_t lLine)
{
   lClAllocs++;
   unsigned char *p = (unsigned char*)malloc(size+2);
   printf("%lxh = alloc(%ld,%s,%d)\n", p, size, pszFile, lLine);
 
   for (int32_t i=0; i<4096 && apMem[i]; i++);
   if (i==4096) { printf("mem overflow\n"); exit(1); }

   apMem[i]  = p;
   alSize[i] = (int32_t)size;

   lMemTotal += size;

   p[0] = 0xEE;
   p[size+1] = 0xEE;

   return p+1;
}

static void operator delete (void *pUserMemory)
{
   if (!pUserMemory)
      return;

   unsigned char *p = (unsigned char*)pUserMemory;

   for (int32_t i=0; i<4096 && (apMem[i]!=(p-1)); i++);
   if (i==4096)
   {
      printf("not found: %lxh at %s %d\n",p,pszGlblMemFile,lGlblMemLine);
      return;
   }

   int32_t lSize = alSize[i];

   if (*(p-1) != 0xEE)
      {  printf("red zone 1 hit: %lxh, %s %d",p,pszGlblMemFile,lGlblMemLine); exit(1); }
   if (*(p+lSize) != 0xEE)
      {  printf("red zone 2 hit: %lxh, %s %d",p,pszGlblMemFile,lGlblMemLine); exit(1); }

   lClFrees++;
   printf("%lxh freed\n", p-1);
   free(p-1);

   lMemTotal -= lSize;
   alSize[i] = 0;
   apMem[i] = 0;
}

#define new new(__FILE__,__LINE__)
#define delete setMemPos(__FILE__,__LINE__),delete

static char *mystrdup(const char *psz,char *pszFile,int32_t lLine)
{
   printf("strdup %s, %d: ",pszFile,lLine);
   char *psz2 = new char[strlen(psz)+1];
   memcpy(psz2,psz,strlen(psz));
   psz2[strlen(psz)]=0;
   return psz2;
}
#define strdup(x) mystrdup(x,__FILE__,__LINE__)

#endif


class BMessageDigest
{
public:
         BMessageDigest ( );
virtual ~BMessageDigest ( );

virtual int32_t
   digestLength   ( );
virtual int32_t
   bufferLength   ( );
unsigned char
  *digest         ( );
virtual void
   reset          ( );
void
   update         (unsigned char c);
void
   update         (unsigned char *pData, int32_t lDataLength);
inline void
   counter        (uint32_t &ulHigh, uint32_t &ulLow);

protected:

virtual void
   generateDigest ( );
virtual void
   transform      ( );
virtual void
   save           ( );
virtual void
   restore        ( );
void
   init           ( );

uint32_t
   lClCountHigh, lClCountLow;
int32_t
   lClBufferCount;
unsigned char
  *pClBuffer;
int32_t
   lClBufferLength;
unsigned char
  *pClDigest;
int32_t
   lClDigestLength;
unsigned char
   ucClBuffer[100];
unsigned char
   ucClDigest[100];
};

inline void BMessageDigest :: counter(uint32_t &ulHigh, uint32_t &ulLow)
   { ulHigh = lClCountHigh; ulLow = lClCountLow; }

class  BMD5
 : public BMessageDigest
{
public:
         BMD5     ( );
virtual ~BMD5     ( );

virtual int32_t
   digestLength   ( );
virtual int32_t
   bufferLength   ( );
virtual void
   reset          ( );

protected:

virtual void
   generateDigest ( );
virtual void
   transform      ( );
virtual void
   save           ( );
virtual void
   restore        ( );

uint32_t
   alClState[4], alClStateCopy[4], alClBuffer[16];
unsigned char
   abClBufferCopy[64];
};

class  BSHA1
 : public BMessageDigest
{

public:
         BSHA1 ( );
virtual ~BSHA1 ( );

virtual int32_t
   digestLength   ( );
virtual int32_t
   bufferLength   ( );
virtual void
   reset          ( );

protected:

virtual void
   generateDigest ( );
virtual void
   transform      ( );
virtual void
   save           ( );
virtual void
   restore        ( );

uint32_t
   alClState[5], alClStateCopy[5], alClBuffer[80];
unsigned char
   abClBufferCopy[64];
};

BMessageDigest :: BMessageDigest ()
{
   pClBuffer = 0;
   pClDigest = 0;

   init ();
}

BMessageDigest :: ~BMessageDigest ()
{
   lClCountHigh = lClCountLow = 0;

   memset (pClBuffer, 0, lClBufferLength);
   memset (pClDigest, 0, lClDigestLength);
}

int32_t BMessageDigest :: digestLength () { return 0; }
int32_t BMessageDigest :: bufferLength () { return 1; }

unsigned char * BMessageDigest :: digest ()
{
   save ();
   int32_t lBufferCount = lClBufferCount;
   uint32_t lCountHigh = lClCountHigh;
   uint32_t lCountLow = lClCountLow;

   generateDigest ();

   restore ();
   lClBufferCount = lBufferCount;
   lClCountHigh = lCountHigh;
   lClCountLow = lCountLow;

   return pClDigest;
}

void BMessageDigest :: reset ()
{
   lClCountHigh = lClCountLow = 0;
   lClBufferCount = 0;
}

void BMessageDigest :: update (unsigned char c)
{
   pClBuffer[lClBufferCount++] = c;

   if (lClBufferCount == lClBufferLength)
   {
      transform ();
      lClBufferCount = 0;
   }

   if (++lClCountLow == 0)
      lClCountHigh++;
}

void BMessageDigest :: update (unsigned char *pData, int32_t lDataLength)
{
   int32_t lCnt = 0;

   if (lClBufferCount != 0)
   {
      int32_t lAvailable = lClBufferLength - lClBufferCount;
      int32_t lMax = lAvailable < lDataLength ? lAvailable : lDataLength;

      for (; lCnt < lMax; lCnt++)
         update (pData[lCnt]);
   }

   int32_t lNumBuffers = (lDataLength - lCnt) / lClBufferLength;

   uint32_t ulNumChars = lNumBuffers * lClBufferLength;

   lClCountLow += ulNumChars;

   if (lClCountLow < ulNumChars)
      lClCountHigh++;

   for (int32_t i = 0; i < lNumBuffers; i++)
   {
      memcpy (pClBuffer, pData + lCnt, lClBufferLength);
      transform ();
      lCnt += lClBufferLength;
   }

   for (; lCnt < lDataLength; lCnt++)
      update (pData[lCnt]);
}

void BMessageDigest :: generateDigest () { }
void BMessageDigest :: transform () { }
void BMessageDigest :: save ()      { }
void BMessageDigest :: restore ()   { }

void BMessageDigest :: init ()
{
   lClBufferLength = bufferLength ();
   lClDigestLength = digestLength ();

   pClBuffer = ucClBuffer;
   pClDigest = ucClDigest;

   reset ();
}

BMD5 :: BMD5 ()
{
   init ();
}

BMD5 :: ~BMD5 ()
{
   memset (alClState, 0, sizeof (alClState));
   memset (alClStateCopy, 0, sizeof (alClStateCopy));
   memset (alClBuffer, 0, sizeof (alClBuffer));
   memset (abClBufferCopy, 0, sizeof (abClBufferCopy));
}

int32_t BMD5 :: digestLength ()  { return 16; }
int32_t BMD5 :: bufferLength ()  { return 64; }

void BMD5 :: reset ()
{
   BMessageDigest :: reset ();

   alClState[0] = 0x67452301;
   alClState[1] = 0xefcdab89;
   alClState[2] = 0x98badcfe;
   alClState[3] = 0x10325476;
}

void BMD5 :: generateDigest ()
{
   uint32_t lLow = lClCountLow << 3;
   uint32_t lHigh = (lClCountLow >> 29) | (lClCountHigh << 3);

   update (128);

   while ((lClCountLow & 63) != 56)
      update (0);

   update ((unsigned char)lLow);
   update ((unsigned char)(lLow >> 8));
   update ((unsigned char)(lLow >> 16));
   update ((unsigned char)(lLow >> 24));
   update ((unsigned char)lHigh);
   update ((unsigned char)(lHigh >> 8));
   update ((unsigned char)(lHigh >> 16));
   update ((unsigned char)(lHigh >> 24));

   unsigned char *pDigest = pClDigest;

   for (int32_t i = 0; i < 4; i++)
   {
      uint32_t ul = alClState[i];

      pDigest[0] = (unsigned char)ul;
      ul >>= 8;
      pDigest[1] = (unsigned char)ul;
      ul >>= 8;
      pDigest[2] = (unsigned char)ul;
      ul >>= 8;
      pDigest[3] = (unsigned char)ul;
      pDigest += 4;
   }
}

void BMD5 :: transform ()
{

#define MD5F(x, y, z)    (z ^ (x & (y ^ z)))
#define MD5G(x, y, z)    (y ^ (z & (x ^ y)))
#define MD5H(x, y, z)    (x ^ y ^ z)
#define MD5I(x, y, z)    (y  ^  (x | ~z))

#define ROTATE(a, n) a = (a << n) | (a >> (32 - n))

#define MD5R0(a, b, c, d, k, s, t) { a += k + t + MD5F (b, c, d); ROTATE (a, s); a += b; }
#define MD5R1(a, b, c, d, k, s, t) { a += k + t + MD5G (b, c, d); ROTATE (a, s); a += b; }
#define MD5R2(a, b, c, d, k, s, t) { a += k + t + MD5H (b, c, d); ROTATE (a, s); a += b; }
#define MD5R3(a, b, c, d, k, s, t) { a += k + t + MD5I (b, c, d); ROTATE (a, s); a += b; }

   int32_t i;
   unsigned char *pBuffer = pClBuffer;

   for (i = 0; i < 16; i++)
   {
      uint32_t ul;

      ul = pBuffer[0];
      ul |= pBuffer[1] << 8;
      ul |= pBuffer[2] << 16;
      ul |= pBuffer[3] << 24;

      pBuffer += 4;
      alClBuffer[i] = ul;
   }

   uint32_t a = alClState[0];
   uint32_t b = alClState[1];
   uint32_t c = alClState[2];
   uint32_t d = alClState[3];

   // Round 0
   MD5R0 (a, b, c, d, alClBuffer[ 0],  7, 0xd76aa478);
   MD5R0 (d, a, b, c, alClBuffer[ 1], 12, 0xe8c7b756);
   MD5R0 (c, d, a, b, alClBuffer[ 2], 17, 0x242070db);
   MD5R0 (b, c, d, a, alClBuffer[ 3], 22, 0xc1bdceee);
   MD5R0 (a, b, c, d, alClBuffer[ 4],  7, 0xf57c0faf);
   MD5R0 (d, a, b, c, alClBuffer[ 5], 12, 0x4787c62a);
   MD5R0 (c, d, a, b, alClBuffer[ 6], 17, 0xa8304613);
   MD5R0 (b, c, d, a, alClBuffer[ 7], 22, 0xfd469501);
   MD5R0 (a, b, c, d, alClBuffer[ 8],  7, 0x698098d8);
   MD5R0 (d, a, b, c, alClBuffer[ 9], 12, 0x8b44f7af);
   MD5R0 (c, d, a, b, alClBuffer[10], 17, 0xffff5bb1);
   MD5R0 (b, c, d, a, alClBuffer[11], 22, 0x895cd7be);
   MD5R0 (a, b, c, d, alClBuffer[12],  7, 0x6b901122);
   MD5R0 (d, a, b, c, alClBuffer[13], 12, 0xfd987193);
   MD5R0 (c, d, a, b, alClBuffer[14], 17, 0xa679438e);
   MD5R0 (b, c, d, a, alClBuffer[15], 22, 0x49b40821);

   // Round 1
   MD5R1 (a, b, c, d, alClBuffer[ 1],  5, 0xf61e2562);
   MD5R1 (d, a, b, c, alClBuffer[ 6],  9, 0xc040b340);
   MD5R1 (c, d, a, b, alClBuffer[11], 14, 0x265e5a51);
   MD5R1 (b, c, d, a, alClBuffer[ 0], 20, 0xe9b6c7aa);
   MD5R1 (a, b, c, d, alClBuffer[ 5],  5, 0xd62f105d);
   MD5R1 (d, a, b, c, alClBuffer[10],  9, 0x02441453);
   MD5R1 (c, d, a, b, alClBuffer[15], 14, 0xd8a1e681);
   MD5R1 (b, c, d, a, alClBuffer[ 4], 20, 0xe7d3fbc8);
   MD5R1 (a, b, c, d, alClBuffer[ 9],  5, 0x21e1cde6);
   MD5R1 (d, a, b, c, alClBuffer[14],  9, 0xc33707d6);
   MD5R1 (c, d, a, b, alClBuffer[ 3], 14, 0xf4d50d87);
   MD5R1 (b, c, d, a, alClBuffer[ 8], 20, 0x455a14ed);
   MD5R1 (a, b, c, d, alClBuffer[13],  5, 0xa9e3e905);
   MD5R1 (d, a, b, c, alClBuffer[ 2],  9, 0xfcefa3f8);
   MD5R1 (c, d, a, b, alClBuffer[ 7], 14, 0x676f02d9);
   MD5R1 (b, c, d, a, alClBuffer[12], 20, 0x8d2a4c8a);

   // Round 2
   MD5R2 (a, b, c, d, alClBuffer[ 5],  4, 0xfffa3942);
   MD5R2 (d, a, b, c, alClBuffer[ 8], 11, 0x8771f681);
   MD5R2 (c, d, a, b, alClBuffer[11], 16, 0x6d9d6122);
   MD5R2 (b, c, d, a, alClBuffer[14], 23, 0xfde5380c);
   MD5R2 (a, b, c, d, alClBuffer[ 1],  4, 0xa4beea44);
   MD5R2 (d, a, b, c, alClBuffer[ 4], 11, 0x4bdecfa9);
   MD5R2 (c, d, a, b, alClBuffer[ 7], 16, 0xf6bb4b60);
   MD5R2 (b, c, d, a, alClBuffer[10], 23, 0xbebfbc70);
   MD5R2 (a, b, c, d, alClBuffer[13],  4, 0x289b7ec6);
   MD5R2 (d, a, b, c, alClBuffer[ 0], 11, 0xeaa127fa);
   MD5R2 (c, d, a, b, alClBuffer[ 3], 16, 0xd4ef3085);
   MD5R2 (b, c, d, a, alClBuffer[ 6], 23, 0x04881d05);
   MD5R2 (a, b, c, d, alClBuffer[ 9],  4, 0xd9d4d039);
   MD5R2 (d, a, b, c, alClBuffer[12], 11, 0xe6db99e5);
   MD5R2 (c, d, a, b, alClBuffer[15], 16, 0x1fa27cf8);
   MD5R2 (b, c, d, a, alClBuffer[ 2], 23, 0xc4ac5665);

   // Round 3
   MD5R3 (a, b, c, d, alClBuffer[ 0],  6, 0xf4292244);
   MD5R3 (d, a, b, c, alClBuffer[ 7], 10, 0x432aff97);
   MD5R3 (c, d, a, b, alClBuffer[14], 15, 0xab9423a7);
   MD5R3 (b, c, d, a, alClBuffer[ 5], 21, 0xfc93a039);
   MD5R3 (a, b, c, d, alClBuffer[12],  6, 0x655b59c3);
   MD5R3 (d, a, b, c, alClBuffer[ 3], 10, 0x8f0ccc92);
   MD5R3 (c, d, a, b, alClBuffer[10], 15, 0xffeff47d);
   MD5R3 (b, c, d, a, alClBuffer[ 1], 21, 0x85845dd1);
   MD5R3 (a, b, c, d, alClBuffer[ 8],  6, 0x6fa87e4f);
   MD5R3 (d, a, b, c, alClBuffer[15], 10, 0xfe2ce6e0);
   MD5R3 (c, d, a, b, alClBuffer[ 6], 15, 0xa3014314);
   MD5R3 (b, c, d, a, alClBuffer[13], 21, 0x4e0811a1);
   MD5R3 (a, b, c, d, alClBuffer[ 4],  6, 0xf7537e82);
   MD5R3 (d, a, b, c, alClBuffer[11], 10, 0xbd3af235);
   MD5R3 (c, d, a, b, alClBuffer[ 2], 15, 0x2ad7d2bb);
   MD5R3 (b, c, d, a, alClBuffer[ 9], 21, 0xeb86d391);

   alClState[0] += a;
   alClState[1] += b;
   alClState[2] += c;
   alClState[3] += d;
}

void BMD5 :: save ()
{
   memcpy (alClStateCopy, alClState, sizeof (alClState));
   memcpy (abClBufferCopy, pClBuffer, sizeof (abClBufferCopy));
}

void BMD5 :: restore ()
{
   memcpy (alClState, alClStateCopy, sizeof (alClState));
   memcpy (pClBuffer, abClBufferCopy, sizeof (abClBufferCopy));
}

BSHA1 :: BSHA1 ()
{
   init ();
}

BSHA1 :: ~BSHA1 ()
{
   memset (alClState, 0, sizeof (alClState));
   memset (alClStateCopy, 0, sizeof (alClStateCopy));
   memset (alClBuffer, 0, sizeof (alClBuffer));
   memset (abClBufferCopy, 0, sizeof (abClBufferCopy));
}

int32_t BSHA1 :: digestLength () { return 20; }
int32_t BSHA1 :: bufferLength () { return 64; }

void BSHA1 :: reset ()
{
   BMessageDigest :: reset ();

   alClState[0] = 0x67452301;
   alClState[1] = 0xefcdab89;
   alClState[2] = 0x98badcfe;
   alClState[3] = 0x10325476;
   alClState[4] = 0xc3d2e1f0;
}

void BSHA1 :: generateDigest ()
{
   uint32_t lLow = lClCountLow << 3;
   uint32_t lHigh = (lClCountLow >> 29) | (lClCountHigh << 3);

   update (128);

   while ((lClCountLow & 63) != 56)
      update (0);

   update ((unsigned char)(lHigh >> 24));
   update ((unsigned char)(lHigh >> 16));
   update ((unsigned char)(lHigh >> 8));
   update ((unsigned char)lHigh);
   update ((unsigned char)(lLow >> 24));
   update ((unsigned char)(lLow >> 16));
   update ((unsigned char)(lLow >> 8));
   update ((unsigned char)lLow);

   unsigned char *pDigest = pClDigest;

   for (int32_t i = 0; i < 5; i++)
   {
      uint32_t ul = alClState[i];

      pDigest[3] = (unsigned char)ul;
      ul >>= 8;
      pDigest[2] = (unsigned char)ul;
      ul >>= 8;
      pDigest[1] = (unsigned char)ul;
      ul >>= 8;
      pDigest[0] = (unsigned char)ul;
      pDigest += 4;
   }
}

void BSHA1 :: transform ()
{
   int32_t i;
   unsigned char *pcBuffer = pClBuffer;
   uint32_t *plBuffer = alClBuffer;

   for (i = 0; i < 16; i++)
   {
      uint32_t ul;

      ul = pcBuffer[0];
      ul <<= 8;
      ul |= pcBuffer[1];
      ul <<= 8;
      ul |= pcBuffer[2];
      ul <<= 8;
      ul |= pcBuffer[3];
      pcBuffer += 4;
      *plBuffer++ = ul;
   }

   for (i = 16; i < 80; i++)
   {
      uint32_t k = plBuffer[-3] ^ plBuffer[-8] ^ plBuffer[-14] ^ plBuffer[-16];
      *plBuffer++ = k << 1 | k >> 31;
   }

   uint32_t s0 = alClState[0];
   uint32_t s1 = alClState[1];
   uint32_t s2 = alClState[2];
   uint32_t s3 = alClState[3];
   uint32_t s4 = alClState[4];

   plBuffer = alClBuffer;

   for (i = 0; i < 20; i++)
   {
      int32_t help = (s0 << 5 | s0 >> 27) + s4 + *plBuffer++;
      help += (((s2 ^ s3) & s1) ^ s3) + 0x5a827999;
      s4 = s3;
      s3 = s2;
      s2 = s1 << 30 | s1 >> 2;
      s1 = s0;
      s0 = help;
   }

   for (i = 0; i < 20; i++)
   {
      int32_t help = (s0 << 5 | s0 >> 27) + s4 + *plBuffer++;
      help += (s1 ^ s2 ^ s3) + 0x6ed9eba1;
      s4 = s3;
      s3 = s2;
      s2 = s1 << 30 | s1 >> 2;
      s1 = s0;
      s0 = help;
   }

   for (i = 0; i < 20; i++)
   {
      int32_t help = (s0 << 5 | s0 >> 27) + s4 + *plBuffer++;
      help += ((s1 & s2) | ((s1 | s2) & s3)) + 0x8f1bbcdc;
      s4 = s3;
      s3 = s2;
      s2 = s1 << 30 | s1 >> 2;
      s1 = s0;
      s0 = help;
   }

   for (i = 0; i < 20; i++)
   {
      int32_t help = (s0 << 5 | s0 >> 27) + s4 + *plBuffer++;
      help += (s1 ^ s2 ^ s3) + 0xca62c1d6;
      s4 = s3;
      s3 = s2;
      s2 = s1 << 30 | s1 >> 2;
      s1 = s0;
      s0 = help;
   }

   alClState[0] += s0;
   alClState[1] += s1;
   alClState[2] += s2;
   alClState[3] += s3;
   alClState[4] += s4;
}

void BSHA1 :: save ()
{
   memcpy (alClStateCopy, alClState, sizeof (alClState));
   memcpy (abClBufferCopy, pClBuffer, sizeof (abClBufferCopy));
}

void BSHA1 :: restore ()
{
   memcpy (alClState, alClStateCopy, sizeof (alClState));
   memcpy (pClBuffer, abClBufferCopy, sizeof (abClBufferCopy));
}


class BSymmetricCipher
{

public:

enum CipherMode { ECB, CBC };

    BSymmetricCipher ( );
virtual
   ~BSymmetricCipher ( );

virtual int32_t
   blockLength       ( );
virtual int32_t
   keyLength         ( );
virtual int32_t
   isStreamCipher    ( );
virtual void
   setKey            (unsigned char *pKey);
virtual void
   encrypt           (unsigned char *pData, int32_t lDataLength);
virtual void
   decrypt           (unsigned char *pData, int32_t lDataLength);
void
   setMode           (CipherMode mode);
CipherMode
   mode              ( );
void
   setCBCInitVector  (unsigned char *pIV);

protected:

virtual void
   encryptBlock      (unsigned char *pBlock);
virtual void
   decryptBlock      (unsigned char *pBlock);
virtual void
   streamCrypt       (unsigned char *pData, int32_t lLength);
void
   init              ( );

CipherMode
   clMode;
int32_t
   lClBlockLength;
unsigned char *
   pClCBCBlock1;
unsigned char *
   pClCBCBlock2;
};

inline void BSymmetricCipher :: setMode (CipherMode mode) { clMode = mode; }
inline BSymmetricCipher::CipherMode BSymmetricCipher :: mode ( ) { return clMode; }

BSymmetricCipher :: BSymmetricCipher ()
{
   pClCBCBlock1 = 0;
   pClCBCBlock2 = 0;

   init ();
}

BSymmetricCipher :: ~BSymmetricCipher ()
{
   memset (pClCBCBlock1, 0, lClBlockLength);
   delete [] pClCBCBlock1;

   memset (pClCBCBlock2, 0, lClBlockLength);
   delete [] pClCBCBlock2;
}

int32_t BSymmetricCipher :: blockLength ()       { return 1; }
int32_t BSymmetricCipher :: keyLength ()         { return 0; }
int32_t BSymmetricCipher :: isStreamCipher () { return 1; }
void BSymmetricCipher :: setKey (unsigned char *pKey) { }

void BSymmetricCipher :: encrypt (unsigned char *pData, int32_t lDataLength)
{
   int32_t i,j;

   if (isStreamCipher ())
   {
      streamCrypt (pData, lDataLength);
   }
   else
   {
      switch (clMode)
      {
         case ECB:
            for (i = 0; i < lDataLength; i += lClBlockLength)
            {
               encryptBlock (pData);
               pData += lClBlockLength;
            }
         break;

         case CBC:
            for (i = 0; i < lDataLength; i += lClBlockLength)
            {
               for (j = 0; j < lClBlockLength; j++)
                  pData[j] ^= pClCBCBlock1[j];

               encryptBlock (pData);

               for (j = 0; j < lClBlockLength; j++)
                  pClCBCBlock1[j] = pData[j];

               pData += lClBlockLength;
            }
         break;
      }
   }
}

void BSymmetricCipher :: decrypt (unsigned char *pData, int32_t lDataLength)
{
   int32_t i,j;

   if (isStreamCipher ())
   {
      streamCrypt (pData, lDataLength);
   }
   else
   {
      switch (clMode)
      {
         case ECB:
            for (i = 0; i < lDataLength; i += lClBlockLength)
            {
               decryptBlock (pData);
               pData += lClBlockLength;
            }
         break;

         case CBC:
            for (i = 0; i < lDataLength; i += lClBlockLength)
            {
               for (j = 0; j < lClBlockLength; j++)
                  pClCBCBlock2[j] = pData[j];

               decryptBlock (pData);

               for (j = 0; j < lClBlockLength; j++)
                  pData[j] ^= pClCBCBlock1[j];

               for (j = 0; j < lClBlockLength; j++)
                  pClCBCBlock1[j] = pClCBCBlock2[j];

               pData += lClBlockLength;
            }
         break;
      }
   }
}

void BSymmetricCipher :: setCBCInitVector (unsigned char *pIV)
{
   for (int32_t i = 0; i < lClBlockLength; i++)
      pClCBCBlock1[i] = pIV[i];
}

void BSymmetricCipher :: encryptBlock (unsigned char *pBlock) { return; }
void BSymmetricCipher :: decryptBlock (unsigned char *pBlock) { return; }
void BSymmetricCipher :: streamCrypt (unsigned char *pData, int32_t lLength) { return; }

void BSymmetricCipher :: init ()
{
   lClBlockLength = blockLength ();

   delete [] pClCBCBlock1;
   pClCBCBlock1 = new unsigned char[lClBlockLength];
   delete [] pClCBCBlock2;
   pClCBCBlock2 = new unsigned char[lClBlockLength];
}

#ifdef WITH_BC4
class BC4
 : public BSymmetricCipher
{
public:
    BC4              ( );
virtual
   ~BC4              ( );

virtual int32_t
   keyLength         ( );
virtual void
   setKey            (unsigned char *pKey);

// protected:
virtual void
   streamCrypt       (unsigned char *pData, int32_t lLength);

int32_t
   lClKeyLength;
unsigned char
   lClX, lClY;
unsigned char
   abClKey[256];
};

BC4 :: BC4 ()
{
   lClX = 0;
   lClY = 0;
   lClKeyLength = 16;
   init ();
}

BC4 :: ~BC4 ()
{
   memset (abClKey, 0, 256);
   lClX = 0;
   lClY = 0;
}

int32_t BC4 :: keyLength ()
{
   return lClKeyLength;
}

void BC4 :: setKey (unsigned char *pKey)
{
   int32_t i;

   for (i = 0; i < 256; i++)
      abClKey[i] = (unsigned char)i;

   int32_t a = 0;
   int32_t b = 0;

   for (i = 0; i < 256; i++)
   {
      b = (pKey[a] + abClKey[i] + b) & 0xff;
      unsigned char help = abClKey[i];
      abClKey[i] = abClKey[b];
      abClKey[b] = help;
      a++;
      a %= lClKeyLength;
   }
}

void BC4 :: streamCrypt (unsigned char *pData, int32_t lLength)
{
   for (int32_t i = 0; i < lLength; i++)
   {
      lClX++;
      lClX &= 0xff;
      int32_t help1 = abClKey[lClX];
      lClY += help1;
      lClY &= 0xff;
      unsigned char help2 = abClKey[lClY];
      abClKey[lClY] = (unsigned char)help1;
      help1 += help2;
      help1 &= 0xff;
      abClKey[lClX] = help2;
      pData[i] ^= abClKey[help1];
   }
}
#endif


num getCurrentTime();

#define STATE_SIZE 1023

class BRG   // random generator
{
public:
   BRG   ( );

void
   seed           ( );
void
   seed           (unsigned char *pabData, int32_t length);

void
   getRandomData  (unsigned char *abData, int32_t length);
unsigned char
   getRandomByte  ( );
int32_t
   getRandomLong  ( );

void
   runSeeder ();
void
   runIncreaser ();

BMD5
   clMD5;
unsigned char
  *abClMD5Digest;

unsigned char
   abClState   [STATE_SIZE+100];
int32_t
   lClStateSize;
volatile int32_t
   lClPos;              // current position in state pool
volatile int32_t
   lClNum;              /* number of random entropy of the pool
                           (all data from seed but max STATE_SIZE) */
volatile int32_t
   lClReadPos;          // lower index into state pool
volatile int32_t
   lClIncLoopCounter;   // internal increment counter
};

BRG :: BRG()
{
   abClMD5Digest     = 0;
   lClPos            = 0;
   lClNum            = 0;
   lClReadPos        = 0;
   lClIncLoopCounter = 0;

   memset (abClState, 0, STATE_SIZE);
   lClStateSize = STATE_SIZE;
}

typedef union
{
   void *p;
   unsigned char ab[8];

}  tmixpab;

void BRG :: seed ()
{
   // ByteArray abSeedData (16);
   unsigned char abSeedData[16];
   // no init. take random stack data by intention.

   // seed with time
   int32_t millis = (int32_t)getCurrentTime();

   // offset 0 to 3: current time millis
   int32_t i;
   for (i = 0; i < 4; i++)
      abSeedData[i] = (unsigned char)(millis >> (i*8));

   // seed with PID
   int32_t currentPID = getpid();

   // offset 4 to 5: current pid
   for (i = 0; i < 2; i++)
      abSeedData[i + 4] = (unsigned char)(currentPID >> (i*8));

   // offset 6 to 13: mem address of abSeedData
   tmixpab o;
   o.p = &abSeedData[0];
   for (i = 0; i < 8; i++)
      abSeedData[i + 6] = o.ab[i];

   // offset 14, 15 are kept as is (random stack data)

   seed (abSeedData, 16);

   for (i = 0; i < 16; i++)
      abSeedData[i] = (unsigned char)0;
}

void BRG :: seed (unsigned char *ptr, int32_t lSize)
{
   lClReadPos = 8;

   int32_t lPos = lClPos;
   int32_t lNum = lClNum;

   lClPos += lSize;

   if (lClPos > STATE_SIZE)
   {
      lClPos %= STATE_SIZE;
      lClNum = STATE_SIZE;
   }
   else
   if (lClNum < STATE_SIZE && lClPos > lClNum)
      lClNum = lClPos;

   for (int32_t i = 0; i < lSize; i += 16)
   {
      int32_t j = lSize - i;

      if (j > 16)
         j = 16;

      if (!abClMD5Digest)
         abClMD5Digest = clMD5.digest ();

      clMD5.update (abClMD5Digest, clMD5.digestLength ());
      int k = lPos + j - STATE_SIZE;

      if (k > 0)
      {
         clMD5.update (abClState + lPos, j - k);
         clMD5.update (abClState, k);
      }
      else
      {
         clMD5.update (abClState + lPos, j);
      }

      clMD5.update (ptr + i, j);
      abClMD5Digest = clMD5.digest ();
 
      for (k = 0; k < j; k++)
      {
         if (lPos >= STATE_SIZE)
            lPos = 0;

         abClState[lPos++] ^= abClMD5Digest[k];
      }
   }
}

unsigned char BRG :: getRandomByte ()
{
   if (lClReadPos < 8)
   {
      return abClMD5Digest[8 + lClReadPos++];
   }

   int32_t lPos = lClPos;
   int32_t lNum = lClNum;
   lClPos += 8;

   if (lClPos > lClNum)
      lClPos %= lClNum;

   for (int32_t i = 0; i < 8; i++)
   {
      clMD5.update (abClMD5Digest + 8, 8);

      int k = lPos + 8 - lNum;

      if (k > 0)
      {
         clMD5.update (abClState + lPos, 8 - k);
         clMD5.update (abClState, k);
      }
      else
      {
         clMD5.update (abClState + lPos, 8);
      }

      abClMD5Digest = clMD5.digest ();

      for (int32_t j = 0; j < 8; j++)
      {
         if (lPos >= lNum)
            lPos = 0;

         abClState[lPos++] ^= abClMD5Digest[j];
      }
   }

   lClReadPos = 1;

   return abClMD5Digest[8];
}

int32_t BRG :: getRandomLong ()
{
   int32_t ret = 0;

   for (int32_t l = 0; l < 8; l++)
      ret = (ret << 8) | ((int32_t)getRandomByte() & 0xff);

   return ret;
}

void BRG :: getRandomData (unsigned char *abData, int32_t length)
{
   for (int32_t l = 0; l < length; l++)
      abData[l] = getRandomByte();
}

// #define _ printf("[%d]\n",__LINE__);

#define uchar unsigned char

static const char *pszBlank78 =
   "                                                                              ";

/*************************************************
 ***        tool functions                     ***
 *************************************************/

static char *mynumtostr(num n, int nDigits, char *pszBuf, int nRadix)
{
   static char szBuf[100];
   if (!pszBuf)
        pszBuf = szBuf;

   #ifdef _WIN32
   if (nRadix == 10)
      sprintf(pszBuf, "%0*I64d", nDigits, n);
   else
      sprintf(pszBuf, "%0*I64X", nDigits, n);
   return pszBuf;
   #else
   if (nRadix == 10)
      sprintf(pszBuf, "%0*lld", nDigits, n);
   else
      sprintf(pszBuf, "%0*llX", nDigits, n);
   return pszBuf;
   #endif
}

static char *mynumtoa(num n, int nDigits=1, char *pszBuf=0) {
   return mynumtostr(n, nDigits, pszBuf, 10);
}

static void hexDump(const char *psz,void *p,int32_t l)
{
   unsigned char *up = (unsigned char*)p;
   printf("%s: ",psz);
   for (int32_t i=0; i<l; i++)
      printf("%02X",up[i]);
   printf("\n");
}

#if defined(__BR_NT__) || defined(__BR_OS2__)
 #include <io.h>
 #include <conio.h>
#endif

void doSleep
 (
   uint32_t msec    //  the delay time in millisecond
 )
{
  #ifdef __BR_NT__
  Sleep(msec);
  #elif defined (__BR_OS2__)
  DosSleep(msec);
  #elif defined (__BR_SUN__) || defined(__BR_HPUX__) || defined(__BR_LINUX__)
  int32_t lSec = msec / 1000;
  int32_t nSec = 1000000 * (msec-1000*lSec); 
  timespec ltime;
  ltime.tv_sec  = lSec;
  ltime.tv_nsec = nSec;
  timespec lRelicTime;  
  lRelicTime.tv_sec  = 0;
  lRelicTime.tv_nsec = 0;  
  nanosleep(&ltime,&lRelicTime);
  #elif defined (__BR_OS390__)
  usleep(msec*1000);
  #elif defined(__BR_BEOS__)
  sleep(msec/1000);  // found nothing better so far
  #else
  // uint32_t ldummy = msec / 1000;  // ldummy = 0 if msec < 1000
  // timestruc_t ltime = { ldummy ,1000000 *(msec-1000*ldummy) };
  // timestruc_t lRelicTime = { 0,0 };
  // int rc;
  // rc = nsleep( &ltime , &lRelicTime);
  int imsec = (int)msec;
  const timespec ts = { imsec / 1000, imsec % 1000 * 1000000 };
  nanosleep(&ts, NULL);
  #endif
}

#ifndef INCLUDE_BCRYPT
num getCurrentTime()
{
   #ifdef __BR_SUN__
   struct timeval tv;
   gettimeofday(&tv,NULL);
   return ((num)(tv.tv_sec * 1000 + tv.tv_usec / 1000));
   #elif defined(__BR_NT__)
   return GetCurrentTime();
   #else
   // timeb currentTime;
   // ftime( &currentTime );
   // return (currentTime.time *1000 + currentTime.millitm);
   struct timeval tv;
   gettimeofday(&tv, NULL);
   return ((num)tv.tv_sec) * 1000 + ((num)tv.tv_usec) / 1000;
   #endif
}
#endif // INCLUDE_BCRYPT

static char upcase
 (
   char c
 )
{
   int32_t lValue = c;
   switch (lValue)
   {
      case 0x84: return (0x8E);
      case 0x94: return (0x99);
      case 0x81: return (0x9A);
      default:   return (toupper(c));
   }
}

uint32_t buildSum(unsigned char *p, int32_t lSize)
{
   uint32_t ulSum = 0;
   for (; lSize-- > 0;)
   {
      unsigned char uc = *p++;
      if (ulSum & (1UL << 31))
          ulSum = (ulSum << 1) | 1;
      else
          ulSum =  ulSum << 1;
      ulSum ^= (uint32_t)uc;
   }
   return ulSum;
}

#if defined(__BR_IX__)

#include <termios.h>
int32_t peekKeyboard()
{
   struct termios saveTerminal;

   if (tcgetattr(STDIN_FILENO, &saveTerminal) < 0)
      return -1;

   struct termios terminal;

   terminal = saveTerminal;
   terminal.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
   terminal.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
   terminal.c_cflag &= ~(CSIZE | PARENB);
   terminal.c_cflag |= CS8;
   terminal.c_cc[VMIN] = 0;
   terminal.c_cc[VTIME]= 5;

   char cKey = 0;

   if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &terminal) < 0)
      return -1;

   int32_t lBytes = read(STDIN_FILENO, &cKey, 1);

   if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &saveTerminal) < 0)
      return -1;

   if (lBytes > 0)
      cKey &= 0xff;

   return cKey;
}
#elif defined(__BR_OS2__)
int32_t peekKeyboard()
{
   if (_kbhit())
      return _getch();

   return -1;
}
#else
int32_t peekKeyboard()
{
   return 0;
}
#endif

/*********************************************************
 ***        file processing part                       ***
 *********************************************************/

#define BLOCK_MAX 30004 // must be 8-padded size plus 4

class BCrypt
{
public:
   BCrypt   ( );
  ~BCrypt   ( );

char
   *mapFilename      (char *pszInfile, int32_t *bEncrypt);
int32_t
    scanInfile       ( );
int32_t
    waitForKey       ( );
int32_t
    readHiddenLine   (char *pszBuf, int32_t maxlen, char *psz2, int32_t lDoComment);
void
    init             ( ),
    log              (const char *format, ...);
int32_t
    parseCommandLine (int32_t argc, char *argv[], int32_t *lRC);
int32_t
    enterKey         ( );
int32_t
    buildTargetFilename ( );
int32_t
    createCipher     ( );
int32_t
    process          ( );
void
    dumpGUIHeader    ( );
void
    buildAbsoluteOutfileName  ( );
int32_t
    queryOverwrite   (char *pszFilename);
int32_t
    docrypt (unsigned char  *inOut,   //  input, and result
             int32_t           *length,  /*  length of the message,
                                          it must be multiple of 8 bytes */
             int32_t        bEncrypt     /*  flag indicates encryption
                                          or decryption            */
             );
BSymmetricCipher
   &cipher           ( );

void
   mathHash (unsigned char *pMD5Hash,   // 16 bytes
             unsigned char *pSHAHash,   // 20 bytes
             unsigned char *pTarget,    // 16 bytes
             int32_t lOperation            // xor, plus, minus, etc.
            );
int32_t
   keyFromSecret  (
      unsigned char *pSecret,       // passphrase
      int32_t           lSecretSize,   // passphrase length
      unsigned char *pKey,          // the target key area
      int32_t           lKeySize,      // between 16 and 32 bytes
      int32_t       *pIterations,      // if non-zero, do exactly so many iterations,
      int32_t        lMaxIter,         // but not more than so many,
      int32_t        lIterMaxTime      // and use just max. so much time
      );
int32_t
   setPhase1Keys     (int iMaxEncryptIterationTime=0);
int32_t
   setPhase2Keys     (unsigned char *abp2Rnd, int32_t lSize);
int32_t
   setInfileName     (char *pszRaw);
void
   putBytes (unsigned char *p, int32_t l, int32_t idx=-1);
void
   putShort (uint32_t l, int32_t idx=-1);
void
   putLong  (uint32_t l, int32_t idx=-1);
int32_t
   writeBuffer       (uchar *pOptData=0, int iOptLen=0);
int32_t
   readBuffer        (int32_t lHowMuch, int32_t lOffset=0);
int
   freadmem          (uchar *pOutBuf,int iBlockSize,int iMaxLen,FILE *pInFile);
uint32_t
   getLong           (int32_t idx);
BRG
   &rg               ( );
void
   setCipherKeyAndIV (unsigned char *pKey, unsigned char *pIV);
int32_t
   tellProgress      (char *pszHead, num lDone, num lTodo, num lStarted);
int32_t
   forceEncrypt      (int32_t bEncrypt);
int32_t
   errorOpenFile     (char *pszFilename, const char *pszFiletype, int iFromInfo);
int32_t
   eraseSingleFile   (char *pszFile, num &lBytes);
#ifdef __BR_NT__
int32_t
   eraseMultipleFiles (char *pszRaw, int32_t &lFiles, num &lBytes);
int32_t
   eraseDirectoryTree (char *pszPath, int32_t lLevel, int32_t &lFiles, int32_t &lDirs, num &lBytes);
#endif

struct
   {
      char  *fullname;  // e.g. /usr/tmp/test.tar.gz
      char  *path;      // e.g. /usr/tmp/
      char  *base;      // e.g.          test
      char  *ext;       // e.g.               tar.gz
      char  *sext;      // scanned extension, if any
   }
   infile;
char  
   *pszClOutfile;

num
   lClInFileSize;

FILE
   *pClInfile,
   *pClOutfile;

int encryptMemory(uchar *pin, int nin, char *pszKey);
int decryptMemory(uchar *pin, int nin, char *pszKey);
uchar *outData ( );
int    outSize ( );

int setMemBlock(uchar *pin, int nin);

uchar
   *pClInMem,
   *pClOutMem;
int
   iClInMemOff,
   iClInMemMax,
   iClOutMemOff,
   iClOutMemMax;

int32_t
   bClEncrypt,
   bClModeForced,
   bClGUIMode,
   bClUseShortExtensions,  // encrypt to "test.brx" i/o "test.doc.brx"
   bClShowCRC,
   bClVerbose,
   bClForceOverwrite,
   lClQuiet,
   bClAskAtEnd;

int32_t
   lClCipher,
   lClTest,          // 1==test outer, 2==test inner
   lClHashFormat;    // 0==comfort, 1==md5 compact, 2==sha compact, 3==both compact

int32_t
   lClInFormat,
   lClInHeadSize,
   lClInVersion,
   lClInAlgoID,
   lClInRelNameSize,
   lClInKeyIterations;

BSymmetricCipher
  *pClCipher;
BRG
  *pClRG;

char  
   *pszClKey;

unsigned char
   abClKeyBuf        [100],
   abClKeyBuf2       [100],
   abClKeyBufSalted  [400];

unsigned char
   abIn        [BLOCK_MAX+1024],
   abOut       [BLOCK_MAX+1024];

int32_t
   lClOutIndex;

unsigned char
   abClKey1Raw [100],   // as large as abClKeyBuf
   abClKey1MD  [32],    // upto 32 bytes. using 16, 24 or 32
   abClIV1     [8],
   abClKey2MD  [32],    // upto 32 bytes. using 16, 24 or 32
   abClIV2     [8],
   abClSalt    [200];   // passphrase salt

int32_t
   lClKeyIterations,
   lClSaltSize;
};

inline BSymmetricCipher &BCrypt :: cipher ( ) { return *pClCipher; }
inline BRG &BCrypt :: rg ( ) { return *pClRG; }

BCrypt :: BCrypt ( )
{
   // i know this is strange, but it saves codesize:
   memset(this, 0, sizeof(*this));

   init(); // for base values init

   pClRG = new BRG();
}

BCrypt :: ~BCrypt ( )
{
   init(); // free all except RG

   delete pClRG;
}

void BCrypt :: init ( )
{
   // called in constructor.
   // called when reusing object e.g. to decrypt.

   BRG *pSaveRG = pClRG;

   if (pClCipher)    delete pClCipher;
   if (infile.fullname) delete [] infile.fullname;
   if (infile.path)  delete [] infile.path;
   if (infile.base)  delete [] infile.base;
   if (infile.ext)   delete [] infile.ext;
   if (infile.sext)  delete [] infile.sext;
   if (pszClOutfile) delete [] pszClOutfile;
   if (pClInfile)    fclose(pClInfile);
   if (pClOutfile)   fclose(pClOutfile);
   if (pClOutMem)    delete [] pClOutMem;

   // wipe-out key remains:
   memset(this, 0, sizeof(*this));

   pClRG = pSaveRG;

   // base inits
   bClShowCRC  = 1;
   lClSaltSize = 16; // by default
}

void BCrypt :: log(const char *format, ...)
{
   if (bClVerbose)
   {
      va_list arg_ptr;
      va_start(arg_ptr, format);
   
      char szTmpBuf[4096];
      vsprintf(szTmpBuf, format, arg_ptr);
   
      printf("%s\n", szTmpBuf);
   }
}

void BCrypt :: setCipherKeyAndIV
 (
   unsigned char *pKey,
   unsigned char *pIV
 )
{
   cipher().setKey(pKey);
   cipher().setCBCInitVector(pIV);
   if (bClVerbose)
   {
      hexDump("set key",pKey,24);
      hexDump("set iv",pIV,8);
   }
}

int32_t BCrypt :: setInfileName (char *pszRaw)
{
   // e.g. /usr/tmp/misc.stuff/../other/x.tar.gz.brx.dat

   #if defined(__BR_NT__) || defined(__BR_OS2__)
   char *pszRawbase = strrchr(pszRaw, '\\');
   if (!pszRawbase)
        pszRawbase  = strchr(pszRaw, ':');
   #else
   char *pszRawbase = strrchr(pszRaw, '/');
   #endif

   #ifdef __BR_NT__
   WIN32_FIND_DATA myfdat;
   HANDLE myfdh = FindFirstFile(pszRaw, &myfdat);
   if (myfdh == INVALID_HANDLE_VALUE)
      {  printf("input file %s not found\n", pszRaw); return 1; }
   // uint32_t ulSizeHi = myfdat.nFileSizeHigh;
   // uint32_t ulSizeLo = myfdat.nFileSizeLow;
   // if (ulSizeLo > (1UL<<31) || ulSizeHi)
   //   {  printf("input file %s too large (> 2 GB), cannot process."); }
   char *pszRel = strdup(myfdat.cFileName);
   if (FindNextFile(myfdh, &myfdat))
      {
         FindClose(myfdh);
         printf("ERROR: filename %s ambiguous\n", pszRaw);
         return 0;
      }
   FindClose(myfdh);
   // we now hold a full relative filename in pszRel.
   if (pszRawbase)
   {
      // append it to the original raw path.
      pszRawbase++;
      int32_t lPathLen = pszRawbase-pszRaw;
      char *psz = new char[lPathLen+strlen(pszRel)+10];
      sprintf(psz, "%.*s%s", (int)lPathLen,pszRaw,pszRel);
      pszRaw = psz;
   }
   else
      // take it as new raw name.
      pszRaw = pszRel;
   #else
   FILE *fTmp = fopen(pszRaw, "rb");
   if (!fTmp)
      {  printf("input file %s not found\n", pszRaw); return 0; }
   fclose(fTmp);
   #endif

   infile.fullname = strdup(pszRaw);

   #if defined(__BR_NT__) || defined(__BR_OS2__)
   char *pszInbase = strrchr(pszRaw, '\\');
   if (!pszInbase)
        pszInbase   = strchr(pszRaw, ':');
   #else
   char *pszInbase  = strrchr(pszRaw, '/');
   #endif

   if (pszInbase)
   {
      // we have a path.
      pszInbase++;
      infile.path  = strdup(pszRaw);
      infile.path[pszInbase-pszRaw] = 0;
      infile.base  = strdup(pszInbase);
   }
   else
   {
      // we have no path.
      infile.path  = strdup("");
      infile.base  = strdup(pszRaw);
   }

   char *pszExt = strchr(infile.base, '.');
   if (pszExt)
   {
      // we have a filename extension.
      infile.ext = strdup(pszExt);
      *pszExt    = 0; // terminate basename
   }
   else
      // we have no filename extension.
      infile.ext = strdup("");

   /*
      now, the input file is specified completely by
      infile.fullname
      infile.path
      infile.base
      infile.ext
   */

   log("infile.fullname %s", infile.fullname);
   log("       path     %s", infile.path);
   log("       base     %s", infile.base);
   log("       ext      %s", infile.ext);

   return 0;
}

int32_t BCrypt :: createCipher()
{
#ifdef WITH_IDEA
   if (lClCipher==CIPHER_IDEA)
   {  pClCipher = new BIdea; return 0; }
#endif
#ifdef WITH_3DES
   if (lClCipher==CIPHER_3DES)
   {  pClCipher = new B3DESede; return 0; }
#endif
#ifdef WITH_BC4
   if (lClCipher==CIPHER_BC4)
   {  pClCipher = new BC4; return 0; }
#endif
   return 1;
}

int32_t BCrypt :: waitForKey ( )
{
   int32_t akt;
   while (1)
   {
      #ifdef __BR_NT__
         akt=getch();
      #elif defined(__BR_OS2__)
         akt=_getch();
      #else
         akt=peekKeyboard();
      #endif
      if (akt>0)
         break;
      doSleep (10);
   }
   return akt;
}

// rc: 0==failed, 1==ok, 2==ESC
int32_t BCrypt ::  queryOverwrite (char *pszFilename)
{
   rg().seed(); // seed with I/O system timing
   FILE *pTmp = fopen(pszFilename, "rb");
   rg().seed();
   if (!pTmp)
      return 1;
   fclose(pTmp);

   if (bClForceOverwrite)
      return 1;

   printf("replace %s? [y]es, [n]o:", pszFilename);
   fflush(stdout);
   int lKey = waitForKey();
   printf("%c\n", lKey);
   #ifdef __BR_NT__
   if (lKey == 27)
      return 2;
   #endif
   if (lKey != 'y')
      return 0;
   return 1;
}

static const char *pszValidChars = "!\"%&/()=?\\+*#,.;:-<>| _[]{}";
// above chars are all used in the "C" language.
// a system not supporting them is a dead system.
// NOT valid are: , accents, umlauts, smileys and the like.

static int32_t validCharacter(int32_t c)
{
   // we must block umlauts etc. - we're never sure
   // if they are inter-system compatible (os/390 etc.)
   if (isalnum(c))
      return 1;
   if (strchr(pszValidChars,c))
      return 1;
   return 0;
}

int32_t BCrypt :: readHiddenLine
 (
   char *line,       // target buffer
   int32_t  maxlen,     // target buffer size
   char *ptoconfirm, /* optional: if non-zero, this is the address of
                        a source pattern. the string to enter should
                        be identical. where it's not, # instead of *
                        is shown to identify typos immediately. */
   int32_t  lDoComment  // use comment about passphrase length
 )
{
   printf(">"); fflush(stdout);

   int32_t pos = 0;
   int32_t weiter = 1;
   int32_t lTellState = 0;
   int32_t l=0;

   while(weiter)
   {
      int akt = waitForKey();

      rg().seed(); // seed with timing of typing

      switch (akt)
      {
       #ifdef __BR_OS390__
       case 22:   // backspace
       #else
       case 8:    // backspace
       #endif
         if(pos>0)
         {
            pos--;
            printf("%c %c",akt,akt); // delete the last char on the screen
            fflush(stdout);
         }
         break;

       case '\n':
       case '\r':
         line[pos]=0; // EOS
         weiter=0;
         break;

       #ifdef __BR_NT__
       case 27:
         printf("\r%s\n", pszBlank78);
         return 1;
       #endif

       default:
         if (!validCharacter(akt))
         {
            if (!lTellState)
            {
               printf("\rvalid characters are: a-e A-E 0-9 %s   \n",pszValidChars);
               lTellState=1;
               printf(">");
               for (l=0;l<pos;l++) printf("*");
            }
            printf(" [ignored invalid character]\n>");
            for (l=0;l<pos;l++) printf("*");
            fflush(stdout);
            break;
         }
         if(pos+1 < maxlen)
         {
            line[pos] = akt;
            char cToPrint = '*';
            if (ptoconfirm != 0)
            {
               char c2 = line[pos];
               #ifdef __BR_OS390__
               __etoa_l(&c2, 1);
               #endif
               if (c2 != ptoconfirm[pos])
                  cToPrint = '#';
            }
            pos++;
            printf("%c", cToPrint);
            fflush(stdout);
         }
         break;
      }
   }

   #ifdef __BR_OS390__
   __etoa_l(line, strlen(line));
   #endif

   printf("< ");
   if (bClShowCRC && lDoComment)
   {
      uint32_t lCRC = 0;
      unsigned char *psz = (unsigned char*)line;
      unsigned char c = 0;
      while (c = *psz++)
         lCRC = ((lCRC << 1) ^ c);
      lCRC = (lCRC % 9) + 1;
      for (uint32_t l=0; l<lCRC; l++)
         printf(".%s", (l%3==2)?" ":"");
   }

   // utter remarks about the passphrase length
   if (lDoComment)
   switch (pos)
   {
      case 0: break;
      case 1:
      case 2:
      case 3:
         printf("   (unsecure)");
         break;
      case 4:
         printf("   (very weak)");
         break;
      case 5:
         printf("   (weak)");
         break;
   };

   printf("\n");
   return 0;
}

int32_t BCrypt :: parseCommandLine
 (
   int32_t  argc,
   char *argv[],
   int32_t *lRC
 )
{
   *lRC = 1; // just in case

   // bcrypt [function] [options] [-k key] infile [outfile]
   //    functions:  e,d,t,T,c,cs,C
   //    options:    f,q,q2

   char *pszInfile = 0;

   int32_t lState = 0;
   int32_t lFiles=0, lDirs=0;
   num  lBytes=0;

   for (int32_t i=1; i<argc; i++)
   {
      char *pArg = argv[i];

      switch (lState)
      {
         case 0:

            // function, or option, or -k, or infile name

            if (!strcmp(pArg, "-e"))
            {
               bClEncrypt = 1;
               bClModeForced = 1;
               continue;
            }
            if (!strcmp(pArg, "-d"))
            {
               bClEncrypt = 0;
               bClModeForced = 1;
               continue;
            }
            if (!strcmp(pArg, "-t"))
            {
               bClEncrypt = 0;
               lClTest = 1;
               bClModeForced = 1;
               continue;
            }
            if (!strcmp(pArg, "-T"))
            {
               bClEncrypt = 0;
               lClTest = 2;
               bClModeForced = 1;
               continue;
            }
            if (   !strcmp(pArg, "-c")
                || !strcmp(pArg, "-cs")
                || !strcmp(pArg, "-C")
               )
            {
               bClEncrypt = 1;
               lClTest = 1;
               bClModeForced = 1;

               if (strlen(pArg)==3)
                  lClHashFormat = 2;   // just sha, compact
               else
               if (pArg[1] == 'C')
                  lClHashFormat = 3;   // md5+sha, compact
               else
                  lClHashFormat = 1;   // just md5, compact

               continue;
            }
            if (   !strcmp(pArg, "-remove")
                || !strcmp(pArg, "-rm")
               )
            {
               lState = 90;
               continue;
            }
            #ifdef __BR_NT__
            if (!strcmp(pArg, "-rmrf"))
            {
               lState = 91;
               continue;
            }
            #endif

            // -bc4 or -3des or -k or infile may come
            if (!strcmp(pArg, "-bc4"))
            {
               lClCipher = CIPHER_BC4;
               continue;
            }

            #ifdef WITH_3DES
            if (!strcmp(pArg, "-3des"))
            {
               lClCipher = CIPHER_3DES;
               lState = 10; // "-k" may come
               continue;
            }
            #endif

            if (!strcmp(pArg, "-f"))
            {
               bClForceOverwrite = 1;
               continue;
            }

            if (!strcmp(pArg, "-q"))
            {
               lClQuiet = 1;
               continue;
            }

            if (!strcmp(pArg, "-q2"))
            {
               lClQuiet = 2;
               continue;
            }

            if (!strcmp(pArg, "-a"))
            {
               bClAskAtEnd = 1;
               continue;
            }

            // fall-through to check for -k

         case 10:
            if (!strcmp(pArg, "-k"))
            {
               lState = 2; // key's next
               continue;
            }

            // fall-through to fetch infile name

         case 3:
            pszInfile = pArg;
            lState = 4; // outfile may come
            continue;

         case 2:
            // this is the key
            pszClKey = pArg;  // best don't copy
            #ifdef __BR_OS390__
            __etoa_l(pszClKey, strlen(pszClKey));
            #endif
            lState = 3; // infile comes
            continue;

         case 4:
            // must be outfile
            pszClOutfile = strdup(pArg);
            lState = -1; // nothing else
            continue;
   
         case 90:
         {
            // file erase mode:
            // treat every parameter as a file, or file pattern, to delete
            if (pArg[0] == '-')
            {
               printf("parameter \"%s\" unexpected after \"-remove\"\n", pArg);
               return 1;
            }

            #ifdef __BR_NT__
            if (eraseMultipleFiles(pArg, lFiles, lBytes))
               return 1;
            #else
            if (eraseSingleFile(pArg, lBytes))
               return 1;
            lFiles++;
            #endif

            continue;
         }

         #ifdef __BR_NT__
         case 91:
         {
            // recursively erase directory tree.
            // we expect exactly one parameter, the directory name.
            if (eraseDirectoryTree(pArg, 0, lFiles, lDirs, lBytes))
               return 1;
            continue;
         }
         #endif

         default:
            printf("unexpected parameter: \"%s\"\n", pArg);
            return 1;
      }
   }

   if (lState==90 && !lClQuiet)
      printf("removed %d files with %lld bytes\n", lFiles, lBytes);

   if (lState==91 && !lClQuiet)
      printf("removed %d files and %d directories with %lld bytes\n",
         lFiles, lDirs, lBytes);

   if (lState >= 90)
   {
      // returning with program rc 0
      *lRC = 0;
      return 1;
   }

   if (pszInfile)
      if (setInfileName(pszInfile))
         return 1;

   if (argc == 2)
      bClGUIMode = 1;

   if (   !pszClOutfile
       &&  bClModeForced
       && !infile.fullname
      )
   {
      printf("no input filename given.\n");
      return 1;
   }

   return (*lRC = 0); // return zero
}

void BCrypt :: buildAbsoluteOutfileName ( )
{
   // place output file in same path as input,
   // IF the output filename doesn't already
   // contain a path.

   #if defined(__BR_NT__) || defined(__BR_OS2__)
   if (   strchr(pszClOutfile, '\\')
       || strchr(pszClOutfile, ':'))
      return;
   #else
   if (   strchr(pszClOutfile, '/'))
      return;
   #endif

   int32_t lSize = strlen(infile.fullname)+strlen(pszClOutfile);
   char *psz  = new char[lSize+20];
   sprintf(psz, "%s%s",
      infile.path, pszClOutfile);
   delete [] pszClOutfile;
   pszClOutfile = psz;
}

// rc: 0==failed, 1==ok, 2==ESC
int32_t BCrypt :: enterKey ( )
{
   abClKeyBuf[0] = 0;

   for (;;)
   {
      printf("please enter %s key, it will not be displayed:\n", bClEncrypt ? "encryption" : "decryption");

      if (readHiddenLine((char*)abClKeyBuf, sizeof(abClKeyBuf)-10, 0, 1))
         return 2; // ESC

      if (!abClKeyBuf[0])
      {
         printf("exiting...\n");
         return 0;
      }

      if (!bClEncrypt)
         break;
      else
      {
         printf("Please type the key again for confirmation.\n"
                "matching characters are displayed as '*', unmatching as '#'\n"
               );
         abClKeyBuf2[0] = 0;
         readHiddenLine((char*)abClKeyBuf2, sizeof(abClKeyBuf2)-10, (char*)abClKeyBuf, 0);

         if (!strlen((char*)abClKeyBuf2))
         {  
            printf("exiting...\n");
            return 0;
            // printf("retry, or enter nothing to exit.\n");
            // continue;
         }

         if (!strcmp((char*)abClKeyBuf,(char*)abClKeyBuf2))
            break;

         printf("mismatch - try again, or enter an empty key to exit.\n");
      }
   }

   pszClKey = (char*)abClKeyBuf;

   if (!abClKeyBuf[0])
      return 0;
   return 1;
}

unsigned char abHeaderNew[] = {
   0x42, 0x43, 0x52, 0x59, // 'B','C','R','Y',
   0x50, 0x54, 0x2F, 0x31  // 'P','T','/','1'
   };

unsigned char abHeaderTrunc[] = {
   0x0A, 0x0D, 0x0A, 0x0A  // LF CR LF LF
   };

unsigned char abBlockHeader[] = {
   0x42, 0x4C, 0x4F, 0x4B  // 'B','L','O','K',
   };

unsigned char abP2Header[] = {
   0x42, 0x52, 0x4E, 0x44  // 'B','R','N','D',
   };  

static void longToFlat(uint32_t l, unsigned char abOut[], int32_t &lOffset)
{
   abOut[lOffset+0] = (unsigned char)((l >> 24) & 0xFFUL);
   abOut[lOffset+1] = (unsigned char)((l >> 16) & 0xFFUL);
   abOut[lOffset+2] = (unsigned char)((l >>  8) & 0xFFUL);
   abOut[lOffset+3] = (unsigned char)((l      ) & 0xFFUL);
   lOffset += 4;
}

static void shortToFlat(uint32_t l, unsigned char abOut[], int32_t &lOffset)
{
   abOut[lOffset+0] = (unsigned char)((l >>  8) & 0xFFUL);
   abOut[lOffset+1] = (unsigned char)((l      ) & 0xFFUL);
   lOffset += 2;
}

void BCrypt :: putBytes(unsigned char *p, int32_t l, int32_t idx)
{
   if (idx != -1)
      memcpy(&abOut[idx],p,l);
   else
   {
      memcpy(&abOut[lClOutIndex],p,l);
      lClOutIndex += l;
   }
}

void BCrypt :: putShort(uint32_t l, int32_t idx)
{
   int32_t lDummy = idx;
   if (lDummy != -1)
      shortToFlat(l, abOut, lDummy);
   else
      shortToFlat(l, abOut, lClOutIndex);
}

void BCrypt :: putLong(uint32_t l, int32_t idx)
{
   int32_t lDummy = idx;
   if (lDummy != -1)
      longToFlat(l, abOut, lDummy);
   else
      longToFlat(l, abOut, lClOutIndex);
}

int32_t BCrypt :: writeBuffer (uchar *pOptData, int iOptLen)
{
   if (pClOutMem) 
   {
      int iremain = iClOutMemMax-iClOutMemOff;
      if (pOptData)
      {
         if (iOptLen>iremain)
            return 10; // buffer overflow
         memcpy(pClOutMem+iClOutMemOff,pOptData,iOptLen);
         iClOutMemOff+=iOptLen;
         return 0;
      }
      else
      {
         if (lClOutIndex>iremain)
            return 10; // buffer overflow
         memcpy(pClOutMem+iClOutMemOff,abOut,lClOutIndex);
         iClOutMemOff+=lClOutIndex;
      }
   }
   else 
   {
      if (!lClTest)
         if (fwrite(abOut,1,lClOutIndex,pClOutfile) != lClOutIndex)
         {  printf("\n\nWRITE ERROR - probably disk full\n"); return 1; }
   }
   lClOutIndex = 0;
   return 0;
}

int32_t BCrypt :: readBuffer (int32_t lHowMuch, int32_t lOffset)
{
   if (pClInMem)
   {
      int ihavemem = iClInMemMax-iClInMemOff;
      int itoread = mymin(ihavemem,lHowMuch);
      if (itoread <= 0)
         return -1;
      memcpy(&abIn[lOffset],pClInMem+iClInMemOff,itoread);
      iClInMemOff+=itoread;
      // printf("READBUF: copied %d, offset %d/%d\n",itoread,iClInMemOff,iClInMemMax);
      return itoread;
   }
   return fread(&abIn[lOffset],1,lHowMuch,pClInfile);
}

int BCrypt :: freadmem(uchar *pOutBuf,int iBlockSize,int iMaxLen,FILE *pInFile)
{
   if (pClInMem)
   {
      int ihavemem = iClInMemMax-iClInMemOff;
      int itoread = mymin(ihavemem,iMaxLen);
      if (itoread <= 0)
         return -1;
      memcpy(pOutBuf,pClInMem+iClInMemOff,itoread);
      iClInMemOff += itoread;
      // printf("READMEM: copied %d, offset %d/%d\n",itoread,iClInMemOff,iClInMemMax);
      return itoread;
   }
   return fread(pOutBuf,iBlockSize,iMaxLen,pClInfile);
}

static uint32_t longFromFlat(unsigned char abIn[], int32_t &lOffset)
{
   uint32_t ulBlock  =
               (((uint32_t)abIn[lOffset+0]) << 24)
            |  (((uint32_t)abIn[lOffset+1]) << 16)
            |  (((uint32_t)abIn[lOffset+2]) <<  8)
            |  (((uint32_t)abIn[lOffset+3])      );
   lOffset += 4;
   return ulBlock;
}

uint32_t BCrypt :: getLong(int32_t idx)
{
   int32_t l1=idx;
   return longFromFlat(abIn,l1); 
}

static unsigned short shortFromFlat(unsigned char abIn[], int32_t &lOffset)
{
   unsigned short ulBlock  =
               (((uint32_t)abIn[lOffset+0]) <<  8)
            |  (((uint32_t)abIn[lOffset+1])      );
   lOffset += 2;
   return ulBlock;
}

int32_t BCrypt :: buildTargetFilename( )
{
   /*
      we have
         infile.fullname
         infile.path
         infile.base
         infile.ext  - extracted extension
         infile.sext - scanned extension, if any
         bClEncrypt
   */

   int32_t lInSize  = strlen(infile.fullname);
   if (infile.ext)
        lInSize += strlen(infile.ext);
   if (infile.sext)
        lInSize += strlen(infile.sext);

   pszClOutfile  = new char[lInSize+20];

   if (bClEncrypt)
   {
      if (bClUseShortExtensions)
         sprintf(pszClOutfile, "%s%s.brx",
            infile.path, infile.base);
      else
         sprintf(pszClOutfile, "%s%s%s.brx",
            infile.path, infile.base, infile.ext);
   }
   else
   {
      if (infile.sext)
         sprintf(pszClOutfile, "%s%s%s",
            infile.path, infile.base, infile.sext);
      else
      {
         // find last ".brx" in extension list, if any
         char *psz = strstr(infile.ext, ".brx");
         char *psz2;
         while (psz && (psz2 = strstr(psz+1, ".brx")))
            psz = psz2;

         // truncate last ".brx", if any
         if (psz)
         {
            sprintf(pszClOutfile, "%s%s%.*s",
               infile.path, infile.base, (int)(psz-infile.ext), infile.ext);
         }
         else
         {
            // no ".brx" in input extension: add ".dat"
            sprintf(pszClOutfile, "%s%s%s.dat",
               infile.path, infile.base, infile.ext);
         }
      }
   }

   return 0;
}

int32_t BCrypt :: forceEncrypt(int32_t bEncrypt)
{
   /*
      allowed is to encrypt an encrypted file (covered in scanInfile()).
      not allowed is to decrypt something not being encrypted.
   */

   if (bClModeForced && (bEncrypt != bClEncrypt))
   {
      printf("error: you want to decrypt, but the input file is not encrypted.\n");
      return 1;
   }

   bClEncrypt = bEncrypt;

   return 0;
}

unsigned char abHeaderOld[] = {
   0x50, 0x41, 0x44  // 'P','A','D'
   };

int32_t BCrypt :: scanInfile ( )
{
   log("scanning \"%s\"", infile.fullname);

   if (!pClInMem)
   {
      #ifdef SFK_W64
   
      struct __stat64 sinfo;
      if (_stat64(infile.fullname, &sinfo))
         return errorOpenFile(infile.fullname, "input", 1);
      if ((lClInFileSize = sinfo.st_size) <= 0)
         {  printf("input file %s seems to have no data\n", infile.fullname); return 1; }
   
      #else
   
      struct stat sinfo;
      if (stat(infile.fullname, &sinfo))
         return errorOpenFile(infile.fullname, "input", 2);
      if ((lClInFileSize = sinfo.st_size) <= 0)
         {  printf("input file %s seems to have no data\n", infile.fullname); return 1; }
   
      #endif
   }

   // did user explicitely select encryption?
   if (bClModeForced && bClEncrypt)
      return 0; // encrypt unconditional

   // open input file for a header scan
   if (!pClInMem)
      if (!(pClInfile = fopen(infile.fullname, "rb")))
         return errorOpenFile(infile.fullname, "input", 3);

   BMD5 md5Head;

   unsigned char abMyHead[8];
   if (freadmem(abMyHead,1,8,pClInfile) < 8)
   {
      // not a bcrypt archive file, header's too short.
      if (pClInfile)
         fclose(pClInfile);
      pClInfile = 0;
      if (forceEncrypt(1))
         return 1;
      return 0;
   }
   md5Head.update(abMyHead,8);

   int32_t bReadExt = 0;

   // for decrypt we have to determine the file format.
   // PAD1 files must be decrypted with plain passphrase key.
   if (!memcmp(abMyHead, abHeaderNew, 8))
   {
      log("[detected PAD3 format]\n");
      lClInFormat = 3;
      bReadExt = 1;
   }
   else
   if (!memcmp(&abMyHead[4], abHeaderOld, 3))
   {
      printf("BCrypt does no int32_ter support files created by earlier tools.\n"
         "use BCrypt V 1.3 for that purpose.\n");
      return 1;
   }
   else
   {
      // not a bcrypt archive file.
      if (pClInfile)
         fclose(pClInfile);
      pClInfile = 0;
      if (forceEncrypt(1))
         return 1;
      return 0;
   }

   bClEncrypt = 0;   // it's an archive - decrypt it

   if (bReadExt)
   {
      memset(abIn,0,sizeof(abIn));

      // read truncation check header
      //    LF    CR    LF    LF
      //    0x0A  0x0D  0x0A  0x0A
      size_t lRead = freadmem(abIn,1,4,pClInfile);
      if (lRead < 4)
         {  printf("error: no trunc header\n"); return 1; }
      if (memcmp(abIn, abHeaderTrunc, 4))
      {
         printf("sorry, but this file is garbage.\n"
            "someone sent it via ftp using ASCII mode,\n"
            "or may have edited it in an ASCII editor.\n"
            "retransmit the file in BINARY mode.\n"
            );
         return 1;
      }
      md5Head.update(abIn,4);

      // read extension header
      lRead = freadmem(abIn,1,4,pClInfile);
      if (lRead < 4)
         {  printf("error: no extension header\n"); return 1; }

      // fetch ext. header length
      int32_t lInIdx = 0;

      lClInHeadSize     = longFromFlat(abIn, lInIdx);

      if (lClInHeadSize > BLOCK_MAX)
         {  printf("error: extension header too large\n"); return 1; }

      lRead = freadmem(&abIn[4],1,lClInHeadSize+16,pClInfile);
      if (lRead < lClInHeadSize+16)
         {  printf("error: incomplete extension header\n"); return 1; }
      md5Head.update(abIn,lClInHeadSize+4);

      lClInVersion      = shortFromFlat(abIn, lInIdx);

      if (lClInVersion < 0x0105)
         {
            printf("this file was encrypted with an older version of bcrypt.\n");
            return 1;
         }

      if (lClInVersion > 0x0107)
         {
            printf("this file was encrypted with a future version of bcrypt.\n");
            return 1;
         }

      // since 0x0103: check hash of header
      if (memcmp(&abIn[4+lClInHeadSize], md5Head.digest(), 16))
         {
            printf("file header corrupted, cannot decrypt.\n");
            return 1;
         }
      log("header hash checked");

      lClInAlgoID       = shortFromFlat(abIn, lInIdx);

      log("version: %lx", lClInVersion);

      // since 0x0102:
      lClInKeyIterations = longFromFlat(abIn, lInIdx);

      log("key iterations: %d", lClInKeyIterations);

      lClCipher = lClInAlgoID;

      // since 0x0107: passphrase salt
      if (lClInVersion >= 0x0107)
      {
         lClSaltSize = shortFromFlat(abIn, lInIdx);
         memcpy(abClSalt, &abIn[lInIdx], lClSaltSize);
         lInIdx += lClSaltSize;
         if (bClVerbose)
            hexDump("passphrase salt", abClSalt, lClSaltSize);
      }

      // read original filename's extension, if supplied at all.
      int32_t lTmpSize = shortFromFlat(abIn, lInIdx);
      if (lTmpSize)
      {
         abIn[lInIdx+lTmpSize] = 0;
         char *psz = strdup((char*)&abIn[lInIdx]);
         #ifdef __BR_OS390__
         __atoe_l(psz, strlen(psz));
         #endif
         char *psz2 = strchr(psz, '.');
         if (psz2)
         {
            infile.sext = strdup(psz2);
            log("scanned original extension \"%s\"", infile.sext);
         }
         delete [] psz;
      }
   }

   #ifdef WITH_IDEA
   if (lClCipher == CIPHER_IDEA) { } else
   #endif
   #ifdef WITH_3DES
   if (lClCipher == CIPHER_3DES) { } else
   #endif
   #ifdef WITH_BC4
   if (lClCipher == CIPHER_BC4) { } else
   #endif
      {
         printf("this file was encrypted with an encryption algorithm (%d)\n"
                "not supported by this version of bcrypt.\n", lClCipher);
         return 1;
      }

   if (forceEncrypt(0))
      return 1;
   return 0;
}

/*
   BRND format:
   1) 4  bytes  "BRND"
   2) 4  bytes  BRND header length  (always 128)
   3) [] bytes  encrypted random noise

   the encrypted random noise from the BRND block
   is md5/sha1 hashed and feeds the symmetric key/IV
   for phase2. all further file BLOK's are encrypted
   using the phase2 keys.

   BLOK format, inner structure:
   1) 4  bytes  "BLOK"
   2) 4  bytes  source block length (netto data)
   3) [] bytes  source block data   (netto data)
   4) 20 bytes  SHA hash field
   5) [] bytes  random padding to 8-byte boundary, 0..7 bytes
   6) 8  bytes  random data. 1st bit of 1st byte is EOF flag.
   the SHA hash is built over 1) 2) 3) 5) 6), skipping 4) itself.

   BLOK format, outer structure:
   1) 4  bytes  "BLOK"
   2) 4  bytes  source block length
   3) [] bytes  encrypted brutto data
   4) 16 bytes  MD5 hash field
   the MD5 hash is built over 1) 2) 3)
*/

int32_t BCrypt :: setPhase1Keys (int iOptMaxTime)
{
   int isubrc = 0;

   if (lClTest != 1)
   {
      if (bClEncrypt || lClInFormat >= 3)
      {
         int32_t lIterations  = 0;
         int32_t lMaxIter     = 0;
         int32_t lIterMaxTime = 0;
   
         if (bClEncrypt)
         {
            if (lClKeyIterations)
            {
               // take environment-selected amount
               lIterations  = lClKeyIterations;
               lMaxIter     = lClKeyIterations;
               lIterMaxTime = 0;
            }
            else
            {
               // take random amount upto a specific time
               lIterations  = MIN_ITERATIONS;   // minimum
               lMaxIter     = 500000; // max for large machines
               lIterMaxTime = iOptMaxTime ? iOptMaxTime : 500; // but just until reaching this
            }
         }
         else
         {
            lIterations    = lClInKeyIterations;
            lMaxIter       = lClInKeyIterations;
            lIterMaxTime   = 0;
         }

         if (bClEncrypt)
            rg().getRandomData(abClSalt, lClSaltSize);

         unsigned char *pSecret = (unsigned char*)pszClKey;
         int32_t lSecSize = strlen(pszClKey);

         if (bClEncrypt || lClInVersion >= 0x0107)
         {
            // since open bcrypt 1.6.0, we append the salt to the passphrase.
            memcpy(abClKeyBufSalted, pSecret, lSecSize);
            memcpy(&abClKeyBufSalted[lSecSize], abClSalt, lClSaltSize);
            pSecret   = abClKeyBufSalted;
            lSecSize += lClSaltSize;
         }

         unsigned char abKeySrc[32];

         // create vast-hashed initial key
         if ((isubrc = keyFromSecret(
            pSecret, lSecSize,
            abKeySrc, 32,
            &lIterations,
            lMaxIter,
            lIterMaxTime
            )))
         {  printf("cannot setup keys\n"); return isubrc; }

         memset(abClKeyBufSalted, 0, sizeof(abClKeyBufSalted));

         // use 32-bytes result material to feed:
         //    24-byte key with IV, or 32-byte key
         memcpy(abClKey1MD, abKeySrc, 32);
         //    8-byte CBC init vector (if used)
         memcpy(abClIV1, &abKeySrc[24], 8);

         lClKeyIterations = lIterations;
      }

      cipher().setMode(BSymmetricCipher::CBC);
      if (!bClEncrypt && lClInFormat==1)
         setCipherKeyAndIV(abClKey1Raw, abClIV1);
      else
         setCipherKeyAndIV(abClKey1MD, abClIV1);

      if (!pClInMem)
      {
         // bc4 is default, say nothing then.
         #ifdef WITH_3DES   
         if (lClCipher==CIPHER_3DES)
            printf("using symmetric cipher tripleDES\n");
         #endif
         #ifdef WITH_IDEA
         if (lClCipher==CIPHER_IDEA)
            printf("using symmetric cipher 3\n");
         #endif
      }

   }  // endif not outer integrity check

   return 0;
}

int32_t BCrypt :: setPhase2Keys(unsigned char *abp2Rnd, int32_t lSize)
{
   unsigned char abKeySrc[32];

   int isubrc = 0;

   int32_t lIterations = 1;

   // reuse keyFromSecret method also for phase2 keys.
   // however vast hashing makes no sense here:
   // one iteration is enough.
   if ((isubrc = keyFromSecret(
      abp2Rnd, lSize,
      abKeySrc, 32,
      &lIterations,
      1, // max iter.
      0  // iter maxtime
      )))
   {  printf("cannot setup phase2 keys\n"); return isubrc; }

   // use 32-bytes result material to feed:
   //    24-byte key and IV, or 32-byte key
   memcpy(abClKey2MD, abKeySrc, 32);
   //    8-byte CBC init vector (if used)
   memcpy(abClIV2, &abKeySrc[24], 8);

   setCipherKeyAndIV(abClKey2MD, abClIV2);

   return 0;
}

int32_t BCrypt :: tellProgress(char *pszHead, num lDone, num lTodo, num lStarted)
{
   if (pClInMem || lClHashFormat || lClQuiet==1)
      return 0;

   int32_t lTime = (getCurrentTime() - lStarted) / 1000;
   if (lTime <= 0) lTime = 1;

   int32_t lDoneK = lDone >> 10;
   int32_t lTodoK = lTodo >> 10;
   int32_t lPerc;
   if (lTodo > 10000)
      lPerc = lDoneK*100/(lTodoK ? lTodoK : 1);
   else
      lPerc = lDone*100/(lTodo ? lTodo : 1);
   int32_t lMax   = 30;
   int32_t lChars = lPerc*lMax/100;
   printf("%s%03d%% >%.*s%.*s< %d k, %d k/sec  \r",
      pszHead,
      lPerc,
      (int)lChars,
      "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
      (int)(lMax-lChars),
      "----------------------------------------",
      lDoneK, lDoneK/lTime);
   fflush(stdout);
   return lPerc;
}

int BCrypt :: setMemBlock(uchar *pin, int nin)
{
   pClInMem = pin;
   iClInMemOff = 0;
   iClInMemMax = nin;

   lClInFileSize = nin;

   if (pClOutMem) delete [] pClOutMem;
   pClOutMem = new uchar[nin+1024+100];
   if (!pClOutMem)
      return 9;

   memset(pClOutMem, 0, nin+1024+100);
   iClOutMemOff = 0;
   iClOutMemMax = nin+1024;

   return 0;
}

/*
   out: rc = 0 OK, else error
   out: outData(), outSize()
*/
int BCrypt :: encryptMemory(uchar *pin, int nin, char *pszKey)
{
   int isubrc = 0;

   rg().seed(); // adds to entropy
   init();
   setMemBlock(pin,nin);
   pszClKey = pszKey;

   bClEncrypt = 1; 
   lClCipher = CIPHER_BC4;
   if ((isubrc=createCipher()))
      return isubrc;
   if ((isubrc=setPhase1Keys(50)))
      return isubrc;
   if ((isubrc=process()))
      return isubrc;

   return 0;
}

uchar *BCrypt :: outData ( ) { return pClOutMem; }
int    BCrypt :: outSize ( ) { return iClOutMemOff; }

int BCrypt :: decryptMemory(uchar *pin, int nin, char *pszKey)
{
   int isubrc = 0;

   // rg().seed();
   init();
   setMemBlock(pin,nin);
   pszClKey = pszKey;

   if ((isubrc=scanInfile())) // sets cipher type, decrypt
      return isubrc;
   if ((isubrc=createCipher()))
      return isubrc;
   if ((isubrc=setPhase1Keys()))
      return isubrc;
   if ((isubrc=process()))
      return isubrc;

   return 0;
}

int32_t BCrypt :: process ( )
{
   if (!pClInMem && !pClInfile)
      if (!(pClInfile = fopen(infile.fullname, "rb")))
         return errorOpenFile(infile.fullname, "input", 4);

   num  lTotalBytes = 0, lTold = 0;
   int32_t lRC = 0, lRead = 0;
   int32_t lTellStep = 8000;
   num lStarted = getCurrentTime();

   BMD5  md5Global;
   BSHA1 shaGlobal;

   char *pszMode;
        if (lClTest && bClEncrypt)  pszMode = (char*)"hashing ";
   else if (lClTest)       pszMode = (char*)"testing ";
   else if (bClEncrypt)    pszMode = (char*)"encrypting ";
   else                    pszMode = (char*)"decrypting ";

   if (bClEncrypt)
   {
      if (!pClInMem && !lClTest)
         if (!(pClOutfile = fopen(pszClOutfile, "wb")))
            return errorOpenFile(pszClOutfile, "output", 5);

      // ENCRYPT a file.
      lClOutIndex = 0;
      putBytes(abHeaderNew,8);

      // truncation check header
      //    LF    CR    LF    LF
      //    0x0A  0x0D  0x0A  0x0A
      // this header gets chopped when using ASCII ftp
      putBytes(abHeaderTrunc,4);

      // header exension structure:
      // offs  len   content
      // ====  ===== ===========
      //    0     4  extension size without md5_hash
      //    4     2  version 0x0107
      // - version 0x0100 fields -    
      //    6     2  algorithm identifier
      //             0x0001 = 3DES
      //             0x0002 = BC4 
      //             0x0003 = IDEA
      // - version 0x0102 fields -    
      //    8     4  key iterations
      //   12     2  passphrase salt size - since version 0x0107
      //   14    []  passphrase salt      - since version 0x0107
      //   []     2  original file extension length or zero
      //   []    []  original file extension (optional)
      // - version 0x0103 fields -
      //   []   16   md5_hash(whole_header_above)
      // - version 0x0104 was skipped,
      //   version 0x0105 uses new (and final) phase2 key setup,
      //   and the ID "BCRYPT/1" i/o "BCRYPT01"
      //   version 0x0106: no change at all, it's just to enforce
      //           upgrade to bcrypt 1.4.1 (with some fixes).

      int32_t lHeadSizeIndex = lClOutIndex;
      putLong(0);                // dummy header size   
      putShort(0x0107);          // version
      putShort(lClCipher);       // algo identifier
      putLong(lClKeyIterations); // key iterations

      // since 0x0107, we add here the passphrase hashing salt.
      // this is added to the entered passphrase for hashing,
      // to disable prehashed table lookups.
      putShort(lClSaltSize);
      putBytes(abClSalt, lClSaltSize);

      // since 0x0106, we no int32_ter put the whole original
      // filename into the open header. if short extensions are
      // selected, we place the original extension, otherwise nothing.
      if (bClUseShortExtensions)
      {
         int32_t lFNESize = strlen(infile.ext);
         putShort(lFNESize);
         putBytes((unsigned char*)infile.ext, lFNESize);
         #ifdef __BR_OS390__
         __etoa_l((char*)&abOut[lClOutIndex-lFNESize], lFNESize);
         #endif
      }
      else
         putShort(0);

      // write extension header:
      int32_t lHeadSize = lClOutIndex-lHeadSizeIndex-4;
      putLong(lHeadSize, lHeadSizeIndex);

      // build header hash:
      BMD5 md5Head;
      md5Head.update(abOut,lClOutIndex);
      putBytes(md5Head.digest(), 16);

      if (writeBuffer())   // resets lClOutIndex
         return 1;

      // create/encrypt/write random noise header,
      // and create phase2 keys from it.
      #define BRND_SIZE (128+4)  // +4 is due to padding of 20 bytes sha hash
      unsigned char abp2Rnd[BRND_SIZE+20];
      rg().getRandomData(abp2Rnd, BRND_SIZE);
      if (bClVerbose)
         hexDump("inner random",abp2Rnd,BRND_SIZE);

      // build inner BRND data:
      //    BRND, length(random+innerHash), random, innerHash
      putBytes(abP2Header,4);
      putLong(BRND_SIZE+20);
      putBytes(abp2Rnd,BRND_SIZE);
      BSHA1 sha;
      sha.update(abOut, lClOutIndex);
      putBytes(sha.digest(),20);
      int32_t l1 = lClOutIndex-8;  // should result in 8-bytes padded length

      if (!lClTest)
         docrypt(&abOut[8],(int32_t*)&l1,1);

      // create outer checksum
      BMD5 md5;
      md5.update(abOut,lClOutIndex);
      putBytes(md5.digest(),16);

      if (writeBuffer())   // resets lClOutIndex
         return 1;

      // switch to phase2 key/IV:
      if (!lClTest && setPhase2Keys(abp2Rnd, BRND_SIZE))
         return 1;

      int32_t bEOF = 0;

      while (!bEOF)
      {
         memset(abIn,0,sizeof(abIn));

         int32_t lWant = BLOCK_MAX;
   
         if ((lRead = readBuffer(lWant)) <= 0)
         {
            printf("error: input file incomplete, size may have changed unexpectedly\n");
            return 1;
         }
   
         log("process %d bytes", lRead);
   
         lTotalBytes += lRead;

         // assert(lTotalBytes <= lClInFileSize);
         if (lTotalBytes > lClInFileSize)
            return 100;

         if (lTotalBytes == lClInFileSize)
            bEOF = 1;

         if (lClTest)
         {
            // just build global hashes over input file:
            md5Global.update(abIn, lRead);
            shaGlobal.update(abIn, lRead);
         }
         else
         {
            // build BLOK header, with source data blocksize,
            // as first 8 bytes of output buffer:
            putBytes(abBlockHeader,4);
            putLong(lRead);
   
            putBytes(abIn,lRead);
   
            int32_t lSHATargetIndex = lClOutIndex;
            lClOutIndex += 20;   // SHA yet to be filled in
   
            int32_t lToCryptLen = lClOutIndex-8;
            if (lToCryptLen % 8)
            {
               // we'll step to the next 8-byte boundary.
               // while doing so, fill the gap with random bytes.
               int32_t lDelta = 8 - (lToCryptLen % 8);
               for (int32_t lFill=0; lFill<lDelta; lFill++)
                  abOut[lClOutIndex++] = rg().getRandomByte();
               log("filled %d gap.",lDelta);
            }
   
            // fill remaining 8 bytes like this:
            //    the very first bit demarks the LAST BLOCK OF FILE.
            //    all further bits are RANDOM.
            unsigned char abFill[8];
            rg().getRandomData(abFill,8);
            if (bEOF)
               abFill[0] |= 0x80; // EOF
            else
               abFill[0] &= 0x7F; // continue
   
            // log("block tail: %lx %lx %lx %lx %lx %lx %lx %lx ",
            //    abFill[0], abFill[1], abFill[2], abFill[3],
            //    abFill[4], abFill[5], abFill[6], abFill[7]
            //    );
   
            putBytes(abFill,8);
   
            // build hash over whole content incl. pad's and randoms,
            // but not including the hash field itself.
            BSHA1 sha;
            sha.update(abOut,lSHATargetIndex);
            // hexDump("hashing",abOut,lSHATargetIndex);
            sha.update(&abOut[lSHATargetIndex+20],lClOutIndex-lSHATargetIndex-20);
            // hexDump("hashing",&abOut[lSHATargetIndex+20],lClOutIndex-lSHATargetIndex-20);
            putBytes(sha.digest(), 20, lSHATargetIndex);
   
            int32_t l1 = lClOutIndex-8;
   
            docrypt(&abOut[8],(int32_t*)&l1,1);
   
            // create outer checksum, without BLOK+blocklength
            BMD5 md5;
            md5.update(abOut,lClOutIndex);
            putBytes(md5.digest(), 16);
   
            if (writeBuffer())
               return 1;
         }

         if (lTotalBytes > lTold + lTellStep)
         {
            tellProgress(pszMode,lTotalBytes,lClInFileSize,lStarted);
            lTold = lTotalBytes;
            if (lTellStep < lClInFileSize / 20)
               lTellStep += lTellStep / 4;
         }
      }
   }
   else
   {
      // DECRYPT or TEST an encrypted file.

      if (lClInFormat >= 3)
      {
         // read/decrypt random noise header
         if (   (lRead = readBuffer(8)) < 8
             || memcmp(abIn,abP2Header,4)
            )
         {  printf("error: no BRND header found\n"); return 1; }

         int32_t lp2HeadSize = getLong(4);
         if (lp2HeadSize != BRND_SIZE+20)
         {  printf("unexpected BRND header size %d\n", lp2HeadSize); return 1; }

         if ((lRead = readBuffer(lp2HeadSize+16,8)) != lp2HeadSize+16)
         {  printf("error while reading BRND header\n"); return 1; }

         // check outer BRND hash
         BMD5 md5;
         md5.update(abIn, 8+lp2HeadSize);
         if (memcmp(&abIn[8+lp2HeadSize], md5.digest(), 16))
            {  printf("outer BRND hash mismatch, file corrupted.\n"); return 1; }
         log("outer BRND hash checked");

         if (lClTest != 1)
         {
            docrypt(&abIn[8],&lp2HeadSize,0);
            if (bClVerbose)
               hexDump("inner random",&abIn[8],lp2HeadSize-20);
   
            // check inner hash
            BSHA1 sha;
            sha.update(abIn,8+lp2HeadSize-20);
            if (memcmp(&abIn[8+lp2HeadSize-20], sha.digest(), 20))
               {  printf("inner BRND hash mismatch, probably wrong key.\n"); return 1; }
            log("inner BRND hash checked");
   
            // switch to phase2 keys/IV:
            if (setPhase2Keys(&abIn[8], BRND_SIZE))
               return 1;
   
            log("checked BRND block - ok");
         }
      }

      if (!pClInMem && !lClTest)
         if (!(pClOutfile = fopen(pszClOutfile, "wb")))
            return errorOpenFile(pszClOutfile, "output", 6);

      unsigned char *pabKey = 0;

      int32_t bEOF = 0;

      // the whole file header was scanned previously.
      // now read and decrypt the blocks.
      while (1)
      {
         if ((lRead = readBuffer(8)) <= 0)
         {
            // all done?
            if (lClInFormat < 3)
               break;   // these had no EOF flag

            if (bEOF || lClTest==1)
            {
               log("\nend of file detected - ok");
               break;
            }
            else
            {
               printf("ERROR: encrypted file got truncated - no EOF flag found\n");
               return 1;
            }
         }

         if (bEOF)
         {
            printf("superfluous data attached to file - ignoring\n");
            break;   // hmmm.
         }

         if (lRead < 8)
            {  printf("error: no block header\n"); return 1; }

         if (memcmp(abIn, abBlockHeader, 4))
            {  printf("error: wrong block header %s\n", abIn); return 1; }

         int32_t lBlock = getLong(4);
         if (lBlock > BLOCK_MAX+20)
            {  printf("illegal block length %d\n", lBlock); return 1; }

         // block is unpadded source length.
         int32_t lPadded = lBlock;
         int32_t lSHAIndex = 8+lPadded;
         int32_t lSHAPostIndex = lSHAIndex+20;

         // above indexes are used only in this case:
         if (lClInFormat >= 3)
            lPadded += 20;

         if (lPadded % 8)
             lPadded = ((lPadded / 8) + 1) * 8;

         // add extra field containing checksum and EOF flag.
         lPadded += 8;

         // add outer hash length?
         int32_t lToRead = lPadded + ((lClInFormat >= 3) ? 16 : 0);
         if ((lRead = readBuffer(lToRead,8)) < lToRead)
            {  printf("error: file incomplete\n"); return 1; }

         lTotalBytes += lBlock;

         if (lClInFormat >= 3)
         {
            // check outer hash
            BMD5 md5;
            md5.update(abIn,8+lPadded);
            if (memcmp(&abIn[8+lPadded], md5.digest(), 16))
               {  printf("outer hash mismatch, file corrupted.\n");
                  return 1; }
            log("outer hash checked");
         }

         if (lClTest != 1) // if not just checking outer
         {
            docrypt(&abIn[8],(int32_t*)&lPadded,0);
   
            if (lClInFormat < 3)
            {
               // rebuild sum just over source data.
               uint32_t ulSum2 = buildSum(&abIn[8],lBlock);
      
               // last 8 bytes of padded area is reserved for checksum.
               uint32_t ulSum1 = getLong(8+lPadded-8); // sic!
               if (ulSum1 != ulSum2)
               {
                  printf("checksum failure (%x %x), probably wrong passphrase\n",ulSum1,ulSum2);
                  return 1;
               }
            }
            else
            {
               // build hash over whole inner data, then compare
               BSHA1 sha;
               sha.update(abIn,lSHAIndex);
               // hexDump("hashing",abIn,lSHAIndex);
               sha.update(&abIn[lSHAPostIndex],8+lPadded-lSHAPostIndex);
               // hexDump("hashing",&abIn[lSHAPostIndex],8+lPadded-lSHAPostIndex);

               if (memcmp(&abIn[lSHAIndex], sha.digest(), 20))
               {
                  printf("inner digest mismatch - probably wrong passphrase\n");
                  return 1;
               }
               log("inner hash checked");
            }
   
            // check for EOF flag         
            if (lClInFormat >= 3 && (abIn[8+lPadded-8+0] & 0x80))
               bEOF = 1;
   
            if (!lClTest)
            {
               if (pClInMem) {
                  if (writeBuffer(&abIn[8],lBlock) != 0)
                     return 10; // buffer overflow
               } else {
               if (fwrite(&abIn[8],1,lBlock,pClOutfile) != lBlock)
               {  printf("\n\nWRITE ERROR - probably disk full\n"); return 1; }
         }
            }
         }
   
         if (lTotalBytes > lTold + lTellStep)
         {
            tellProgress(pszMode,lTotalBytes,lClInFileSize,lStarted);
            lTold = lTotalBytes;
            if (lTellStep < lClInFileSize / 20)
               lTellStep += lTellStep / 4;
         }
      }
   }

   if (bClEncrypt)
   {
      // assert(lTotalBytes==lClInFileSize);
      if (lTotalBytes != lClInFileSize)
         return 101;
   }

   // tell a nice 100% result. in decryption case we have the problem
   // that progress indicator counts netto output against brutto infile size.
   // this still could be perfectionized in the future.
   if (tellProgress(pszMode,lTotalBytes,lTotalBytes,lStarted) > 0)
      printf("\n");

   int32_t lTime = (getCurrentTime() - lStarted) / 1000;
   if (lTime <= 0) lTime = 1;

   if (!lClTest && pClOutfile)
      fclose(pClOutfile);
   if (pClInfile)
   fclose(pClInfile);

   pClOutfile  = 0;
   pClInfile   = 0;

   // pClOutMem is KEPT for further use.

   if (bClEncrypt && lClTest==1)
   {
      unsigned char *pmd5 = md5Global.digest();
      unsigned char *psha = shaGlobal.digest();

      int32_t i; // required by linux gcc

      if (!lClHashFormat)
      {
         printf("%s\nMD5  hash over %s:   (%lld bytes)\n\n     ",
            pszBlank78,infile.fullname,lTotalBytes);
         for (i=0; i<16; i++)
            printf("%02X%s%s",pmd5[i], (i&1)?" ":"", ((i&3)==3)?" ":"");
         printf("\n");
         
         // beware of feature overload... md5 is sufficient for GUI users.
         // printf("\nSHA1 hash over %s:   (%lld bytes)\n\n     ",
         //    infile.fullname,lTotalBytes);
         // for (i=0; i<20; i++)
         //    printf("%02X%s%s",psha[i], (i&1)?" ":"", ((i&3)==3)?" ":"");
         // printf("\n\n");
      }
      else
      {
         // compact format for automatic file checks:
         if (lClHashFormat & 1)
            for (i=0; i<16; i++)
               printf("%02X",pmd5[i]);
         if (lClHashFormat >= 2)
            for (i=0; i<20; i++)
               printf("%02X",psha[i]);
         printf("\n");
      }

      return 0;
   }

   if (pClInMem == 0 && lClQuiet != 1)
    printf("%s\r%s %lld bytes at %lld k/sec\n",
      pszBlank78,
      bClEncrypt ? "encrypted" : "decrypted",
      lTotalBytes,
      (lTotalBytes>>10)/lTime);

   if (lClQuiet)
      return 0;

   if (lClTest==1)
      printf("\n*** outer integrity checked - OK ***\n");
   else
   if (lClTest==2)
      printf("\n*** full integrity checked - OK ***\n");
   else
   if (bClGUIMode)
      if (bClEncrypt)
         printf("\n*** ENCRYPTION DONE. ***\n");
      else
         printf("\n*** DECRYPTION DONE. ***\n");

   return 0;
}

int32_t BCrypt :: docrypt
 (
   unsigned char  *inOut,
   int32_t           *length,
   int32_t            bEncrypt
 )
{
   if (bEncrypt)
      cipher().encrypt(inOut, *length);
   else
      cipher().decrypt(inOut, *length);
   return *length;
}

// loop skeleton, fetching two input bytes
#define HashLoopBegin for(i=0;i<20;i++){uc1=pMD5Hash[i&0xF];uc2=pSHAHash[i];

// loop skeleton end, writing output byte
#define HashLoopEnd   pTarget[i&0xF]^=uc3;}

void BCrypt :: mathHash
 (
   unsigned char *pMD5Hash,   // 16 bytes
   unsigned char *pSHAHash,   // 20 bytes
   unsigned char *pTarget,    // 16 bytes
   int32_t lOperation
 )
{
   unsigned char uc1,uc2,uc3;
   uint32_t ul=0;
   int32_t i;
   
   memset(pTarget, 0, 16);

   // operations supported:
   //    + ^ * | & - -

   switch (lOperation)
   {
      case 0:  // + with carry
         HashLoopBegin;
            ul = uc1 + uc2;
            if (ul > 0xFFUL)
               uc3 = (unsigned char)ul+1;
            else
               uc3 = (unsigned char)ul;
         HashLoopEnd;
         break;

      case 1:  // binary XOR
         HashLoopBegin;
            uc3 = uc1 ^ uc2;
         HashLoopEnd;
         break;

      case 2:  // * churning 16-bit result into 8-bit
         HashLoopBegin;
            ul  = uc1 * uc2;
            uc3 = (unsigned char)((ul >> 8) ^ ul);
         HashLoopEnd;
         break;

      case 3:  // binary OR
         HashLoopBegin;
            uc3 = uc1 | uc2;
         HashLoopEnd;
         break;

      case 4:  // binary AND
         HashLoopBegin;
            uc3 = uc1 & uc2;
         HashLoopEnd;
         break;

      case 5:  // subtract 2nd from 1st,
         HashLoopBegin;
            ul  = uc1 - uc2;  // churning 16-bit result into 8-bit
            uc3 = (unsigned char)((ul >> 8) ^ ul);
         HashLoopEnd;
         break;
      
      case 6:  // subtract 1st from 2nd,
         HashLoopBegin;
            ul  = uc2 - uc1;  // churning 16-bit result into 8-bit
            uc3 = (unsigned char)((ul >> 8) ^ ul);
         HashLoopEnd;
         break;

      default:
         printf("mathHash: illegal operation %d\n", lOperation);
         exit(1);
         break;         
   }
}

#define CHURN_SIZE (16*7)

int32_t BCrypt :: keyFromSecret
 (
   unsigned char *pSecret,
   int32_t           lSecretSize,
   unsigned char *pKey,
   int32_t           lKeySize,   // usually 16 or 24
   int32_t       *pIterations,   // if non-zero, do exactly so many iterations,
   int32_t        lMaxIter,      // but not more than so many,
   int32_t        lIterMaxTime   // and use max. so much time.
 )
{
   // this method has two goals:
   // (1) consume cpu power
   // (2) create a target bit area derived from pSecret,
   //     loosing as few informations as possible

   // assert(*pIterations > 0);
   if (*pIterations <= 0)
      return 110;

   // assert(lMaxIter > 0);
   if (lMaxIter <= 0)
      return 111;

   unsigned char abMD5[16*2];
   unsigned char abSHA[20*2];
   unsigned char abChurn[CHURN_SIZE];

   memset(abMD5, 0, sizeof(abMD5));
   memset(abSHA, 0, sizeof(abSHA));
   memset(abChurn, 0, sizeof(abChurn));

   BMD5 md5;
   BSHA1 sha;

   md5.update(pSecret, lSecretSize);
   memcpy(abMD5, md5.digest(), 16);

   sha.update(pSecret, lSecretSize);
   memcpy(abSHA, sha.digest(), 20);

   num lStarted = getCurrentTime();

   int32_t lIter=0;
   int32_t lOper=0;
   for (;;lIter++)
   {
      if (lIterMaxTime)
      {
         // max-time mode: only used with encryption
         if (   lIter >= *pIterations
             && getCurrentTime() - lStarted >= lIterMaxTime
            )
            break;

         // if someone runs the tool on a large machine,
         // avoid weak target machines to take ages for result decryption:
         if (lIter >= lMaxIter)  // absolute maximum
            break;
      }
      else
      {
         // hashing for a fixed number of iterations
         if (!pClInMem && !lClQuiet && lIter>0 && !(lIter%1000))
         {  printf("hashing... (%d) \r", lIter); fflush(stdout); }

         if (lIter >= *pIterations)
            break;
      }

      for (lOper=0; lOper<7; lOper++)
         mathHash(abMD5, abSHA, &abChurn[lOper*16], lOper);

      if (bClVerbose && !(lIter%1000))
         hexDump("churning",abChurn,CHURN_SIZE);

      md5.update(abChurn, CHURN_SIZE);
      md5.update(pSecret, lSecretSize);
      memcpy(abMD5, md5.digest(), 16);

      sha.update(abChurn, CHURN_SIZE);
      sha.update(pSecret, lSecretSize);
      memcpy(abSHA, sha.digest(), 20);
   }

   if (pClInMem==0 && lIter > 1 && !lClQuiet)
      printf("hashed %d iterations in %d msec\n", lIter, (int32_t)(getCurrentTime() - lStarted));
   // else
   //    printf("hashed phase2 keys\n");

   // write-back info how many iterations were done
   *pIterations = lIter;

   // always one further:
   for (lOper=0; lOper<7; lOper++)
      mathHash(abMD5, abSHA, &abChurn[lOper*16], lOper);

   md5.update(abChurn, CHURN_SIZE);
   md5.update(pSecret, lSecretSize);
   memcpy(&abMD5[16], md5.digest(), 16);

   sha.update(abChurn, CHURN_SIZE);
   sha.update(pSecret, lSecretSize);
   memcpy(&abSHA[20], sha.digest(), 20);

   // fill target key from abMD5, abMD5Second, abSHA, abSHASecond:
   if (lKeySize < 16 || lKeySize > 32)
      {  printf("kfs: illegal keysize %d\n", lKeySize); return 1; }

   memcpy(pKey, abMD5, lKeySize);
   for (int32_t i=0; i<lKeySize; i++)
      pKey[i] ^= abSHA[i];

   // hexDump("keysource", pKey, lKeySize);

   return 0;
}

int32_t BCrypt :: errorOpenFile(char *pszFilename, const char *pszFiletype, int iFromInfo)
{
   printf("error: can't open %s file %s\n", pszFiletype, pszFilename);
   return 1;
}

int32_t BCrypt :: eraseSingleFile
 (
   char  *pszFile,
   num   &lBytes
 )
{
   num lFileSize = 0;

   struct stat sinfo;

   if (stat(pszFile, &sinfo))
      return errorOpenFile(pszFile, "target", 7);

   if ((lFileSize = sinfo.st_size) > 0)
   {
      FILE *f=fopen(pszFile, "r+b");
      if (!f)
         return errorOpenFile(pszFile, "target", 8);
   
      int32_t lNameSize  = strlen(pszFile);
      int32_t lShortSize = lNameSize;
      if (lShortSize > 14)
          lShortSize = 14;
   
      unsigned char abRandom[50000];
      int32_t lMaxBlockSize = sizeof(abRandom);
   
      int32_t bProgressBar = (lFileSize > 1048576);
   
      if (!lClQuiet)
         printf("del %s", pszFile);
   
      unsigned char abVoid[] = { 0x56, 0x4F, 0x49, 0x44 };  // "VOID"
   
      // do this once, takes too much time otherwise
      // rg().getRandomData(abRandom, sizeof(abRandom));
      memset(abRandom, 0xEE, sizeof(abRandom));
   
      num lTotalBytes;
      for (lTotalBytes=0; lTotalBytes<lFileSize;)
      {
         num lRemain = lFileSize-lTotalBytes;
   
         int32_t lSectorSize = sizeof(abRandom);
         if (lSectorSize > lRemain)
             lSectorSize = lRemain;
   
         // fill next sector to write with random data and "VOID" markers
         memcpy(abRandom, abVoid, 4);
         if (lSectorSize >=4 )
            memcpy(abRandom+lSectorSize-4, abVoid, 4);
   
         if (fwrite(abRandom, 1, lSectorSize, f) != lSectorSize)
         {
            printf("\nfailed to erase %s at position %ld\n", pszFile, ftell(f)); 
            break; // no return with rc, it may be more important to delete further files
         }
   
         lTotalBytes += lSectorSize;
      }

      lBytes += lTotalBytes;
   
      if (!lClQuiet)
         printf("\n");
   
      num lLastPos = ftell(f);
      fclose(f);
   
      if (lLastPos != lFileSize)
         printf("\nfailed to erase %s at position %s\n", pszFile, mynumtoa(lLastPos));
   }

   if (remove(pszFile))
      printf("failed to remove %s\n", pszFile);

   return 0;
}

#ifdef __BR_NT__
int32_t BCrypt :: eraseMultipleFiles
 (
   char  *pszRaw,
   int32_t  &lFiles,
   num   &lBytes
 )
{
   #if defined(__BR_NT__) || defined(__BR_OS2__)
   char *pszRawbase = strrchr(pszRaw, '\\');
   if (!pszRawbase)
        pszRawbase  = strchr(pszRaw, ':');
   #else
   char *pszRawbase = strrchr(pszRaw, '/');
   #endif

   if (pszRawbase)
       pszRawbase++;

   WIN32_FIND_DATA myfdat;
   HANDLE myfdh = FindFirstFile(pszRaw, &myfdat);
   if (myfdh == INVALID_HANDLE_VALUE)
      {  printf("file %s not found\n", pszRaw); return 1; }
   
   do
   {
      char *pszRel = myfdat.cFileName;

      if (myfdat.dwFileAttributes & 0x10)
         continue; // directory

      int32_t lSize = myfdat.nFileSizeLow;

      if (pszRawbase)
      {
         // append it to the original raw path.
         int32_t lPathLen = pszRawbase-pszRaw;
         char *psz = new char[lPathLen+strlen(pszRel)+10];
         sprintf(psz, "%.*s%s", (int)lPathLen,pszRaw,pszRel);
         eraseSingleFile(psz, lBytes);
         delete [] psz;
      }
      else
         eraseSingleFile(pszRel, lBytes);

      lFiles++;
   }
   while (FindNextFile(myfdh, &myfdat));

   FindClose(myfdh);  

   return 0;
}

int32_t BCrypt :: eraseDirectoryTree
 (
   char  *pszPath,
   int32_t   lLevel,
   int32_t  &lFiles,
   int32_t  &lDirs,
   num   &lBytes
 )
{
   int32_t lSize1  = strlen(pszPath);
   char *pszPat = new char[lSize1+10];
   sprintf(pszPat, "%s\\*", pszPath);

   WIN32_FIND_DATA myfdat;
   HANDLE myfdh = FindFirstFile(pszPat, &myfdat);
   if (myfdh == INVALID_HANDLE_VALUE)
      {  printf("directory %s not found\n", pszPath); return 1; }

   do
   {
      char *pszFile = myfdat.cFileName;

      if (   !strcmp(pszFile, ".")
          || !strcmp(pszFile, ".."))
         continue;

      int32_t lSize2  = strlen(myfdat.cFileName);
      char *pszSub = new char[lSize1+lSize2+10];
      sprintf(pszSub, "%s\\%s", pszPath, myfdat.cFileName);

      if (myfdat.dwFileAttributes & 0x10)
      {
         // subdirectory
         int32_t lRC = eraseDirectoryTree(pszSub, lLevel+1, lFiles, lDirs, lBytes);
         if (lRC)
         {
            delete [] pszSub;
            return lRC;
         }
      }
      else
      {
         // normal file
         int32_t lSize = myfdat.nFileSizeLow;
         eraseSingleFile(pszSub, lBytes);
         lFiles++;
      }

      delete [] pszSub;
   }
   while (FindNextFile(myfdh, &myfdat));

   FindClose(myfdh);

   if (rmdir(pszPath))
      printf("failed to remove %s\n", pszPath);
   else
      lDirs++;

   return 0;
}
#endif

void BCrypt :: dumpGUIHeader ( )
{
   printf("\n"
   "   Open BCrypt V " VERSTR " - 128 bit file security\n\n");

   if (bClEncrypt)
      printf(
   "      AAA                      42 7C 2A 18\n"
   "    AAAAAAA            =       E7 4E 78 65\n"
   "   AAA   AAA           ===     B3 FA FC 05\n"
   "   AAA   AAA    ===========    D3 03 F4 1F\n"
   "   AAAAAAAAA    ============   47 7C E9 92\n"
   "   AAA   AAA    ===========    6E C0 E2 F2\n"
   "   AAA   AAA           ===     9A 62 0D 93\n"
   "   AAA   AAA           =       BA 05 F3 C1\n"
   "\n"
   "                               ENCRYPTION\n"
   "\n");
   else
      printf(
   "   42 7C 2A 18                    AAA    \n"
   "   E7 4E 78 65         =        AAAAAAA  \n"
   "   B3 FA FC 05         ===     AAA   AAA \n"
   "   D3 03 F4 1F  ===========    AAA   AAA \n"
   "   47 7C E9 92  ============   AAAAAAAAA \n"
   "   6E C0 E2 F2  ===========    AAA   AAA \n"
   "   9A 62 0D 93         ===     AAA   AAA \n"
   "   BA 05 F3 C1         =       AAA   AAA \n"
   "\n"
   "                               DECRYPTION\n"
   "\n");

   printf(
   "       source:  %s\n"
   "       target:  %s\n\n",
   infile.fullname,
   pszClOutfile);
}

#ifdef WITH_SELFTEST
static int32_t singleSelfTest(char *pszMySelf, int32_t lSize, int32_t lDelay)
{
   // create and write a random testfile of size 'lSize'
   unsigned char *p = new unsigned char[lSize];

   int32_t i;
   for (i=0; i<lSize; i++)
      p[i] = (unsigned char)rand();

   BMD5 md;
   md.update(p, lSize);
   unsigned char *pDigest = md.digest();

   FILE *f=fopen("btest.tmp","wb");
   if (!f)
      {  printf("\nunable to open file\n"); return -1; }
   fwrite(p,1,lSize,f);
   fclose(f);
   delete [] p;

   // build a random password.
   char szPass[220];
   const char *pszSrc =
      "abcdefghijklmnopqrstuvwxyz"  // no other way
      "ABCDEFGHIJKLMNOPQRSTUVWXYZ"  // to run this also with os/390,
      "0123456789_";                // as EBCDIC is not contiguous.
   int32_t lSrcLen  = strlen(pszSrc);
   int32_t lPassLen = 5+(rand()%200);
   for (i=0; i<lPassLen; i++)
      szPass[i] = pszSrc[rand()%lSrcLen];
   szPass[i] = 0;

   char szCmd[1024];

   // encrypt test file
   sprintf(szCmd, "%s -e -f -q -k %s btest.tmp btest2.tmp", pszMySelf, szPass);
   system(szCmd);

   // decrypt test file
   sprintf(szCmd, "%s -d -f -q -k %s btest2.tmp btest3.tmp", pszMySelf, szPass);
   system(szCmd);

   // verify test file
   unsigned char *p2 = new unsigned char[lSize];
   f=fopen("btest3.tmp","rb");
   if (!f)
      {  printf("\nunable to open file (2)\n"); return -1; }
   fread(p2,1,lSize,f);
   fclose(f);

   BMD5 md2;
   md2.update(p2, lSize);
   unsigned char *pDigest2 = md2.digest();

   if (memcmp(pDigest, pDigest2, 16))
   {
      printf("\ndigest mismatch,\n"
             "filesize %d, passphrase \"%s\"\n",
             lSize, szPass);
      return -1;
   }

   printf("test OK, filesize %d, phraselen %d, hash ", lSize, lPassLen);
   for (i=0; i<16; i++)
      printf("%02X",pDigest2[i]);
   printf("\n");
   fflush(stdout);

   delete [] p2;

   doSleep(lDelay);

   return 0;
}
#endif

#ifndef INCLUDE_BCRYPT

int main(int argc, char *argv[])
{
   #ifdef MEMTRACE
   memset(alSize, 0, sizeof(alSize));
   memset(apMem, 0, sizeof(apMem));
   #endif

   char *pown = argv[0];

   if (argc > 1 && !strcmp(argv[1], "-memtest"))
   {
     srand(time(0));

     BCrypt b;
     b.rg().seed();

     char *pszKey = (char*)"foo";

     for (int itry=0; itry<1000; itry++)
     {
      uchar abin[100000+100];
      int nin = 100000 - (rand()%50000);
      for (int i=0; i<nin; i++)
         abin[i] = 1+(rand()%255);
      strcpy((char*)abin, "The quick brown fox.");

      // encrypt
      if (b.encryptMemory(abin, nin, pszKey))
         { printf("encrypt failed.\n"); exit(1); }

      uchar *pcrypt = new uchar[b.outSize()];
      int ncrypt = b.outSize();
      memcpy(pcrypt,b.outData(),ncrypt);

      // decrypt
      if (b.decryptMemory(pcrypt, ncrypt, pszKey))
         { printf("decrypt failed.\n"); exit(1); }
 
      if (memcmp(abin,b.outData(),nin)) {
         printf("... output mismatch\n");
         exit(1);
      } else {
         printf("test %d output matches.\r",itry+1);
      }

      delete [] pcrypt;
     }
     return 0;
   }

   #ifdef WITH_SELFTEST
   if (argc > 1 && !strcmp(argv[1], "-selftest"))
   {
      srand(getCurrentTime());
   
      int32_t lSizeToTest = 1;
      int32_t lDelay = 0;
      if (argc > 2)
         lSizeToTest = atol(argv[2]);
      if (argc > 3)
         lDelay = atol(argv[3]);
   
      for (int32_t i=lSizeToTest; i<1048576; i++)
         if (singleSelfTest(argv[0], i, lDelay))
            {  printf("\ntest failed\n"); exit(1);   }
   }
   else
   #endif
   {

   BCrypt b;
   b.rg().seed();

   if (argc < 2)
   {
      printf("Open BCrypt - 128-bit portable file encryption\n"
             "Version " VERSTR " of Jun-2023. Not related to linux bcrypt.\n"
             #ifdef __BR_NT__
             "\nHow to use with WINDOWS EXPLORER:\n"
             "\nto ENCRYPT a file:\n"
             "   take a file with your mouse and drop it on bcrypt.exe\n"
             "\n"
             "to DECRYPT a file:\n"
             "   take the .brx file with your mouse and drop it on bcrypt.exe\n"
             "\n"
             "for even more convenience, you may also:\n"
             "   - associate the file extension .brx with bcrypt.exe\n"
             "   - create a desktop shortcut for bcrypt.exe and drop files on this\n"
             "\npress '2' for command-line infos, and any other key to exit.\n"
             #endif
            );
    #ifdef __BR_NT__
    int32_t lQuick = 0;
    #else
    int32_t lQuick = '2';
    #endif

    while (1)
    {
     int32_t lKey = 0;
     if (lQuick)
     {
         lKey = lQuick;
         lQuick = 0;
     }
     else
         lKey = b.waitForKey();

     switch (lKey)
     {
      case '2':
      printf("\nshort command-line usage:\n\n"
             "   bcrypt [-e|-d] [-k key] infile [outfile]\n"
             "   -e  encrypt     (optional)\n"
             "   -d  decrypt     (optional)\n"
             "   -k  specify key (otherwise type it interactively)\n"
             "\nexamples:\n"
             "   bcrypt -e secret.zip secret.zix\n"
             "   bcrypt securedocs.brx\n"
             "\npress '3' for more information, and any other key to exit.\n"
            );
      break;

      case '3':
      printf("\nthis tool uses bc4 128-bit symmetric encryption.\n"
             "bc4 is an algorithm fully compatible to rc4 (tm).\n"
             "the key is built from a combined md5/sha1 hash of the passphrase you enter.\n"
             "\n"
             "bcrypt does it's best to protect against brute-force and plaintext attacks.\n"
             "however the security of your data relies vitally on the passphrase you select.\n"
             "\nit should\n"
             "- comprise at least 6 characters.\n"
             "  you may also enter complete sentences with blanks.\n"
             "- comprise at least 2 digits or\n"
             "- consist of non-existing words like memplfawzk.\n"
             "- have nothing to do with your friends and relatives.\n"
             "\npress '4' for more, and any other key to exit.\n"
             );
      break;

      case '4':
      printf("\nfilename extensions of encrypted files:\n"
             "\nby default, .brx is added to the existing extension on encryption.\n"
             "this may be changed by setting BCRYPT=shortnames\n"
             "bcrypt will then replace the old filename extension by .brx,\n"
             "and store the original extension in the file header.\n"
             #ifdef __BR_OS390__
             "\nthe os/390 version of bcrypt shows nasty '!' characters on passphrase input.\n"
             "this is a feature per definitionem, and does not influence security.\n"
             #endif
             "\n"
             "\"have you REALLY typed the correct passphrase?\"\n"
             "\n"
             "bcrypt calculates a passphrase checksum from 1 to 10,\n"
             "which is displayed as a row of DOTS behind the entered passphrase.\n"
             "whenever you enter the same passphrase, you see the same number of dots.\n"
             "so, if you doubt that a recipient of a .brx file really typed\n"
             "the passphrase you told, simply ask \"how many dots do you see?\" -\n"
             "and if there's a mismatch with the number of dots YOU see,\n"
             "tell the recipient to check the keyboard SHIFT/LOCK or NUM/LOCK keys.\n"
             "this feature could be turned off by setting BCRYPT=nodots\n"
             "\nyou may combine several settings in the BCRYPT variable like this:\n\n"
             "   set BCRYPT=nodots,shortnames\n"
             "\npress '5' for more, and any other key to exit.\n"
            );
      break;

      case '5':
      printf("\nfull command-line usage:\n"
             "\n  bcrypt [function] [options] [-k key] infile [outfile]\n"
             "\n    functions:  e,d,t,T,c,cs,C\n"
             "      t=test outer integrity of archive (no key required)\n"
             "      T=test inner integrity of archive (key required)\n"
             "      c=calculate md5 cs=sha1 C=md5+sha1\n"
             "\n    options:    f,q,q2,a\n"
             "      f=force target file overwrite\n"
             "      q=quiet mode q2=half-quiet mode\n"
             "      a=ask at end for a key to continue\n"
             "\n  bcrypt -remove|rm filepattern [filepattern2] [...]\n"
             "      try to erase file content, then remove file(s).\n"
             "      will probably not erase content on compressed\n"
             "      or encrypted drives.\n"
             #ifdef __BR_NT__
             "  bcrypt -rmrf path\n"
             "      try to erase whole directory tree content\n"
             #endif
             #ifdef WITH_SELFTEST
             "\n  bcrypt -selftest [initial_file_size] [delay_msec]\n"
             "      create random testfile, encrypt, decrypt, verify\n"
             #endif
             "\n"
             #ifdef WITH_3DES
             "you may use 168-bit TripleDES instead of bc4. to do so,\n"
             "specify option -3des before -k, or set the environmentvariable\n"
             "\n    BCRYPT=3des\n"
             #endif
             "\npress '6' for more, and any other key to exit.\n"
             );
      break;

      case '6':
      printf("\nkey iterations amount:\n"
             "\nto slow down brute force attacks, bcrypt iterates many times\n"
             "over the entered key before using it's hash for en/decryption.\n"
             "the number of times this is done is determined by your machine's\n"
             "CPU power (in encryption case). you may also preselect the amount\n"
             "by setting:\n"
             "\n   set BCRYPT=keyiter:amount\n"
             "\nwhere amount makes sense from %d to 500000.\n"
             "\ncreating a file hash (checksum):\n"
             #ifdef __BR_NT__
             "\ndrop a file on bcrypt.exe, then press ESC, or type in the commandline:\n"
             #endif
             "\n   bcrypt.exe -c infile\n"
             "\nthis creates an md5 hash (checksum) of the file.\n"
             "it could be used to check if two files are identical.\n"
             "\npress any key to exit.\n",
             (int32_t)MIN_ITERATIONS
             );
      break;

      default:
      return 1;
      break;
     }
    }
   }

   b.lClCipher = CIPHER_BC4;  // by default

   char *pszEnv = getenv("BCRYPT");
   if (pszEnv)
   {
      char *psz = 0;
      b.bClVerbose = strstr(pszEnv, "verbose") ? 1 : 0;
      b.bClShowCRC = strstr(pszEnv, "nodots") ? 0 : 1;
      b.bClUseShortExtensions = strstr(pszEnv, "shortnames") ? 1 : 0;
      #ifdef WITH_3DES
      if (strstr(pszEnv, "3des")) b.lClCipher = CIPHER_3DES;
      #endif
      if (psz = strstr(pszEnv, "keyiter:"))
         if ((b.lClKeyIterations = atol(psz+8)) < 100)
            b.lClKeyIterations = 100;
   }

   int32_t lRC=0;

   if (b.parseCommandLine(argc, argv, &lRC))
      return lRC;

   // scan inputfile for size and type,
   // and determine if to en- or decrypt.
   if (b.scanInfile())
      return 1;

   if (!b.lClTest && !b.pszClOutfile)
      if (b.buildTargetFilename())
         return 1;

   // if started from colorful explorer,
   if (!b.lClTest && b.bClGUIMode)
   {
      // to avoid user worries,
      // make sure also outfile has same path as infile
      b.buildAbsoluteOutfileName();

      // bring up a big hello to the user
      b.dumpGUIHeader();
   }

   if (!b.lClTest)
      if (!(lRC = b.queryOverwrite(b.pszClOutfile)))
         return 1;

   // if key not already supplied,
   if (lRC!=2 && !b.pszClKey && b.lClTest!=1)
      // let user now enter it interactively.
      if (!(lRC=b.enterKey()))
         return 1;

   if (lRC==2)
   {
      // user pressed ESC: hash input file
      b.lClTest = 1;
      b.bClEncrypt = 1;
      b.bClModeForced = 1;
      printf("hashing %s...\n\n", b.infile.fullname);
      if (b.pClInfile)
      {
         fclose(b.pClInfile);
         b.pClInfile = 0;
      }
   }

   if (b.lClTest==1 && !b.bClEncrypt && b.lClInFormat < 3)
      {  printf("cannot test this.\n"); return 1; }

   if (b.createCipher())
      {  printf("failed to create cipher\n"); return 1; }

   if (b.lClTest!=1 && b.setPhase1Keys())
      return 1;

   // en/decryption/testing
   if (b.process())
   {
      printf("\n# # # # #  %sCRYPTION FAILED  # # # # #\n",
         b.bClEncrypt ? "EN" : "DE");
      return 1;
   }

   if (!b.pClInMem && (b.bClGUIMode || b.bClAskAtEnd))
   {
      #if defined(__BR_NT__) || defined(__BR_OS2__)
      printf("\npress any key to continue.\n");
      b.waitForKey();
      #else
      printf("\n");
      #endif
   }

   }  // endblock to cleanup BCrypt instance

   #ifdef MEMTRACE
   printf("%d news, %d deletes\n", lClAllocs, lClFrees);
   printf("memory list:\n");
   for (int32_t i=0; i<4096; i++)
      if (apMem[i])
      {
         printf("   %d %lxh\n", alSize[i], apMem[i]);
         hexDump("", apMem[i]+1, alSize[i]);
      }
   #endif

   return 0;
}

#endif // INCLUDE_BCRYPT

